"""Default settings for Richie's simple picture plugin."""
from django.conf import settings

SIMPLEPICTURE_PRESETS = {
    "default": {
        "src": {"size": (500, 500), "crop": "smart"},  # easythumbnail options
        "srcset": [
            {"options": {"size": (1000, 1000), "crop": "smart"}, "descriptor": "1000w"},
            {"options": {"size": (2000, 2000), "crop": "smart"}, "descriptor": "2000w"},
        ],
        "sizes": "100vw",  # e.g 1000px or 100vw
    }
}

SIMPLEPICTURE_PRESETS.update(getattr(settings, "RICHIE_SIMPLEPICTURE_PRESETS", {}))



# Generated by Django 2.2.6 on 2019-10-14 16:01

from django.db import migrations, models

import richie.apps.core.fields.multiselect


class Migration(migrations.Migration):

    dependencies = [("courses", "0008_auto_20191001_1212")]

    operations = [
        migrations.AddField(
            model_name="coursepluginmodel",
            name="variant",
            field=models.CharField(
                blank=True,
                choices=[(None, "Default")],
                help_text="Optional glimpse variant for a custom look.",
                max_length=50,
                null=True,
                verbose_name="variant",
            ),
        )
    ]



"""
SimpleText CMS plugin factories
"""
import factory

from .models import SimpleText


class SimpleTextFactory(factory.django.DjangoModelFactory):
    """
    Factory to create random instances of SimpleText for testing.
    """

    class Meta:
        model = SimpleText

    body = factory.Faker("text", max_nb_chars=42)



# Generated by Django 3.1.3 on 2020-11-02 18:12

from django.db import migrations, models

import richie.apps.core.fields.multiselect


class Migration(migrations.Migration):

    dependencies = [
        ("courses", "0017_auto_20200827_1011"),
    ]

    operations = [
        migrations.AddField(
            model_name="course",
            name="is_listed",
            field=models.BooleanField(
                default=True,
                help_text="Tick if you want the course to be visible on the search page.",
                verbose_name="is listed",
            ),
        ),
        migrations.AlterField(
            model_name="courserun",
            name="languages",
            field=richie.apps.core.fields.multiselect.MultiSelectField(
                choices=[
                    ("af", "Afrikaans"),
                    ("ar", "Arabic"),
                    ("ar-dz", "Algerian Arabic"),
                    ("ast", "Asturian"),
                    ("az", "Azerbaijani"),
                    ("bg", "Bulgarian"),
                    ("be", "Belarusian"),
                    ("bn", "Bengali"),
                    ("br", "Breton"),
                    ("bs", "Bosnian"),
                    ("ca", "Catalan"),
                    ("cs", "Czech"),
                    ("cy", "Welsh"),
                    ("da", "Danish"),
                    ("de", "German"),
                    ("dsb", "Lower Sorbian"),
                    ("el", "Greek"),
                    ("en", "English"),
                    ("en-au", "Australian English"),
                    ("en-gb", "British English"),
                    ("eo", "Esperanto"),
                    ("es", "Spanish"),
                    ("es-ar", "Argentinian Spanish"),
                    ("es-co", "Colombian Spanish"),
                    ("es-mx", "Mexican Spanish"),
                    ("es-ni", "Nicaraguan Spanish"),
                    ("es-ve", "Venezuelan Spanish"),
                    ("et", "Estonian"),
                    ("eu", "Basque"),
                    ("fa", "Persian"),
                    ("fi", "Finnish"),
                    ("fr", "French"),
                    ("fy", "Frisian"),
                    ("ga", "Irish"),
                    ("gd", "Scottish Gaelic"),
                    ("gl", "Galician"),
                    ("he", "Hebrew"),
                    ("hi", "Hindi"),
                    ("hr", "Croatian"),
                    ("hsb", "Upper Sorbian"),
                    ("hu", "Hungarian"),
                    ("hy", "Armenian"),
                    ("ia", "Interlingua"),
                    ("id", "Indonesian"),
                    ("ig", "Igbo"),
                    ("io", "Ido"),
                    ("is", "Icelandic"),
                    ("it", "Italian"),
                    ("ja", "Japanese"),
                    ("ka", "Georgian"),
                    ("kab", "Kabyle"),
                    ("kk", "Kazakh"),
                    ("km", "Khmer"),
                    ("kn", "Kannada"),
                    ("ko", "Korean"),
                    ("ky", "Kyrgyz"),
                    ("lb", "Luxembourgish"),
                    ("lt", "Lithuanian"),
                    ("lv", "Latvian"),
                    ("mk", "Macedonian"),
                    ("ml", "Malayalam"),
                    ("mn", "Mongolian"),
                    ("mr", "Marathi"),
                    ("my", "Burmese"),
                    ("nb", "Norwegian Bokmål"),
                    ("ne", "Nepali"),
                    ("nl", "Dutch"),
                    ("nn", "Norwegian Nynorsk"),
                    ("os", "Ossetic"),
                    ("pa", "Punjabi"),
                    ("pl", "Polish"),
                    ("pt", "Portuguese"),
                    ("pt-br", "Brazilian Portuguese"),
                    ("ro", "Romanian"),
                    ("ru", "Russian"),
                    ("sk", "Slovak"),
                    ("sl", "Slovenian"),
                    ("sq", "Albanian"),
                    ("sr", "Serbian"),
                    ("sr-latn", "Serbian Latin"),
                    ("sv", "Swedish"),
                    ("sw", "Swahili"),
                    ("ta", "Tamil"),
                    ("te", "Telugu"),
                    ("tg", "Tajik"),
                    ("th", "Thai"),
                    ("tk", "Turkmen"),
                    ("tr", "Turkish"),
                    ("tt", "Tatar"),
                    ("udm", "Udmurt"),
                    ("uk", "Ukrainian"),
                    ("ur", "Urdu"),
                    ("uz", "Uzbek"),
                    ("vi", "Vietnamese"),
                    ("zh-hans", "Simplified Chinese"),
                    ("zh-hant", "Traditional Chinese"),
                ],
                help_text="The list of languages in which the course content is available.",
                max_choices=50,
                max_length=255,
            ),
        ),
    ]



"""
Courses application admin
"""
from itertools import chain

from django import forms
from django.conf import settings
from django.contrib import admin
from django.core.exceptions import PermissionDenied, ValidationError
from django.db import models as django_models
from django.db import transaction
from django.http import HttpResponseBadRequest, HttpResponseForbidden, JsonResponse
from django.urls import re_path
from django.utils.decorators import method_decorator
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _
from django.views.decorators.http import require_POST

from cms.admin.placeholderadmin import FrontendEditableAdminMixin
from cms.api import Page
from cms.extensions import PageExtensionAdmin
from cms.utils import page_permissions
from parler.admin import TranslatableAdmin
from parler.forms import TranslatableModelForm

from ..core.admin import link_field
from . import models
from .fields import CourseRunSplitDateTimeField
from .forms import AdminCategoryForm, AdminLicenceForm
from .helpers import snapshot_course
from .widgets import CourseRunSplitDateTimeWidget

REQUIRE_POST = method_decorator(require_POST)


class CategoryAdmin(FrontendEditableAdminMixin, PageExtensionAdmin):
    """Admin class for the Category model"""

    form = AdminCategoryForm


class CourseRunAdminForm(TranslatableModelForm):
    """Admin form used for frontend editing."""

    class Meta:
        model = models.CourseRun
        fields = [
            "direct_course",
            "title",
            "resource_link",
            "start",
            "end",
            "enrollment_start",
            "enrollment_end",
            "languages",
        ]

    def __init__(self, *args, **kwargs):
        """
        If the form is instanciated to update an existing course run:
            > show the direct course select box only if the course has one or more snapshots
              and limit choices to either the master course or one of its snapshots

        If the form is instanciated to create a new course run and the "Add" form is opened via
        the widget from the course detail page on the frontend, the "direct_course" field receives
        the "id" of the related course as initial value (e.g. happens when you pass it via the
        querystring in Django admin):
            > hide the direct course select box
        """
        super().__init__(*args, **kwargs)

        if "direct_course" not in self.fields:
            return

        if self.instance.pk:
            course_query = (
                self.instance.get_course()
                .get_snapshots(include_self=True)
                .filter(extended_object__publisher_is_draft=True)
                .distinct()
            )
        else:
            course_query = models.Course.objects.filter(
                extended_object__publisher_is_draft=True
            )

            # pylint: disable=no-member
            user = self.request.user
            if getattr(settings, "CMS_PERMISSION", False) and not user.is_superuser:
                course_query = models.Course.objects.filter(
                    django_models.Q(extended_object__pagepermission__user=user)
                    | django_models.Q(
                        extended_object__pagepermission__group__user=user
                    ),
                    extended_object__pagepermission__can_change=True,
                    extended_object__publisher_is_draft=True,
                )

            if kwargs.get("initial", {}).get("direct_course"):
                self.fields["direct_course"].widget = forms.HiddenInput()

        self.fields["direct_course"].choices = chain(
            [("", self.fields["direct_course"].empty_label)],
            [(c.pk, c.extended_object.get_title()) for c in course_query],
        )

        if len(course_query) < 2:
            self.fields["direct_course"].widget = forms.HiddenInput()

    def clean_direct_course(self):
        """Ensure that the user has the required permissions to change the related course page."""
        course = self.cleaned_data["direct_course"]
        if (
            course
            and getattr(settings, "CMS_PERMISSION", False)
            and not page_permissions.user_can_change_page(
                self.request.user,  # pylint: disable=no-member
                course.extended_object,
                course.extended_object.node.site,
            )
        ):
            raise ValidationError(
                "You do not have permission to change this course page."
            )

        return course


class CourseRunAdmin(FrontendEditableAdminMixin, TranslatableAdmin):
    """Admin class for the CourseRun model"""

    frontend_editable_fields = (
        "direct_course",
        "title",
        "resource_link",
        "start",
        "end",
        "enrollment_start",
        "enrollment_end",
        "languages",
    )
    list_display = ["id"]
    form = CourseRunAdminForm
    formfield_overrides = {
        django_models.DateTimeField: {
            "form_class": CourseRunSplitDateTimeField,
            "widget": CourseRunSplitDateTimeWidget,
        }
    }

    def has_module_permission(self, request):
        """
        Hide course runs from the admin page as frontend editing provides a better experience.
        """
        return False

    def get_form(self, request, obj=None, **kwargs):
        """Add request to form so we can validate fields according to the user."""
        form = super().get_form(request, obj=obj, **kwargs)
        form.request = request
        return form

    def has_view_permission(self, request, obj=None):
        """Allow view only if the user is allowed to change the related course page."""
        if obj:
            course_page = obj.direct_course.extended_object
            if not course_page.publisher_is_draft:
                return False
            if getattr(
                settings, "CMS_PERMISSION", False
            ) and not page_permissions.user_can_change_page(
                request.user, course_page, course_page.node.site
            ):
                return False
        return super().has_view_permission(request, obj=obj)

    def has_change_permission(self, request, obj=None):
        """Allow change only if the user is allowed to change the related course page."""
        if obj:
            course_page = obj.direct_course.extended_object
            if not course_page.publisher_is_draft:
                return False
            if getattr(
                settings, "CMS_PERMISSION", False
            ) and not page_permissions.user_can_change_page(
                request.user, course_page, course_page.node.site
            ):
                return False
        return super().has_change_permission(request, obj=obj)

    def has_delete_permission(self, request, obj=None):
        """Allow change only if the user is allowed to change the related course page."""
        if obj and getattr(settings, "CMS_PERMISSION", False):
            course_page = obj.direct_course.extended_object
            if not page_permissions.user_can_change_page(
                request.user, course_page, course_page.node.site
            ):
                return False
        return super().has_delete_permission(request, obj=obj)

    # pylint: disable=no-member
    def save(self, commit=True):
        """
        Trigger check that will mark the related course page dirty if its content has changed
        since it was last saved.
        """
        super().save(commit=commit)
        if commit is True:
            self.instance.mark_course_dirty()
        return self.instance


class CourseAdmin(FrontendEditableAdminMixin, PageExtensionAdmin):
    """Admin class for the Course model"""

    list_display = ["title", "is_listed"]
    frontend_editable_fields = (
        "code",
        "duration",
        "effort",
        "is_listed",
        "is_self_paced",
    )

    # pylint: disable=no-self-use
    def title(self, obj):
        """
        Display the course title as a read-only field from the related page
        """
        return obj.extended_object.get_title()

    def get_urls(self):
        """
        Add admin URL to trigger the snapshot of a course.
        """
        url_patterns = super().get_urls()
        return [
            re_path(
                r"^(?P<course_id>[0-9]+)/snapshot/$",
                self.admin_site.admin_view(self.snapshot),
                name="cms_course_snapshot",
            )
        ] + url_patterns

    # pylint: disable=unused-argument
    @REQUIRE_POST
    @transaction.atomic
    def snapshot(self, request, course_id, *args, **kwargs):
        """
        Snapshotting a course is making a copy of the course page with all its permissions and
        extensions, and placing the copy as the first child of the page being snapshotted, then
        moving all the course run of the page being snapshotted as children of the new snapshot
        so we keep record of the course as it was when these course runs were played.
        """
        try:
            page = Page.objects.select_related("node__site").get(
                publisher_is_draft=True, course__id=course_id
            )
        except Page.DoesNotExist:
            return HttpResponseBadRequest(force_str(_("Course could not be found.")))

        try:
            new_page = snapshot_course(page, request.user, simulate_only=False)
        except PermissionDenied as context:
            return HttpResponseForbidden(force_str(context))

        return JsonResponse({"id": new_page.course.id})


class OrganizationAdmin(PageExtensionAdmin):
    """
    Admin class for the Organization model
    """

    list_display = ["title", "code"]

    # pylint: disable=no-self-use
    def title(self, obj):
        """
        Get the page title from the related page
        """
        return obj.extended_object.get_title()


class PageRoleAdmin(admin.ModelAdmin):
    """
    Admin class for the PageRole model.
    """

    fields = [
        "role",
        link_field("page"),
        link_field("group", anchor=_("See user group")),
        link_field("folder", anchor=_("See filer folder")),
    ]
    list_display = ["id", "role", "page"]
    list_filter = ["role"]
    search_fields = ("id", "page__title_set__title")

    def get_fieldsets(self, request, obj=None):
        fieldsets = super().get_fieldsets(request, obj)
        if not obj:
            fieldsets[0][1]["fields"] = ["role", "page"]
        return fieldsets

    def get_readonly_fields(self, request, obj=None):
        """Don't allow modifying the object once created."""
        if obj:
            return self.fields
        return []


class PersonAdmin(PageExtensionAdmin):
    """Admin class for the Person model"""

    list_display = ["title"]

    # pylint: disable=no-self-use
    def title(self, obj):
        """
        Display the person page title as a read-only field from the related page
        """
        return obj.extended_object.get_title()


class LicenceAdmin(TranslatableAdmin):
    """
    Admin class for the Licence model
    """

    list_display = ["name"]
    form = AdminLicenceForm


admin.site.register(models.Category, CategoryAdmin)
admin.site.register(models.Course, CourseAdmin)
admin.site.register(models.CourseRun, CourseRunAdmin)
admin.site.register(models.Licence, LicenceAdmin)
admin.site.register(models.Organization, OrganizationAdmin)
admin.site.register(models.PageRole, PageRoleAdmin)
admin.site.register(models.Person, PersonAdmin)



# Generated by Django 3.1.3 on 2020-11-23 10:46

import django.db.models.deletion
from django.db import migrations, models

import parler


class Migration(migrations.Migration):

    dependencies = [
        ("cms", "0022_auto_20180620_1551"),
        ("courses", "0020_auto_20201118_1153"),
    ]

    operations = [
        # change "on_delete" hook to allow deleting the related CMS pages
        migrations.AlterField(
            model_name="courserun",
            name="extended_object",
            field=models.OneToOneField(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="cms.page",
            ),
        ),
        migrations.AddField(
            model_name="courserun",
            name="draft_course_run",
            field=models.OneToOneField(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="public_course_run",
                to="courses.courserun",
            ),
        ),
        migrations.AddField(
            model_name="courserun",
            name="direct_course",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="runs",
                to="courses.course",
            ),
        ),
        migrations.AlterField(
            model_name="courserun",
            name="resource_link",
            field=models.CharField(
                blank=True, max_length=200, null=True, verbose_name="resource link"
            ),
        ),
        migrations.CreateModel(
            name="CourseRunTranslation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "language_code",
                    models.CharField(
                        db_index=True, max_length=15, verbose_name="Language"
                    ),
                ),
                ("title", models.CharField(max_length=255, verbose_name="title")),
                (
                    "master",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="translations",
                        to="courses.courserun",
                    ),
                ),
            ],
            options={
                "verbose_name": "Course run translation",
                "db_table": "richie_course_run_translation",
                "unique_together": {("language_code", "master")},
            },
            bases=(parler.models.TranslatedFieldsModelMixin, models.Model),
        ),
    ]



# Generated by Django 3.1.3 on 2020-11-23 10:46

import django.db.models.deletion
from django.db import migrations, models


def populate_direct_course(apps, schema_editor):
    """
    Populate the "direct_course" field with the information we currently have in the
    node tree, then delete all course run releated pages.
    """
    Page = apps.get_model("cms", "Page")
    Course = apps.get_model("courses", "Course")
    CourseRun = apps.get_model("courses", "CourseRun")
    CourseRunTranslation = apps.get_model("courses", "CourseRunTranslation")
    Title = apps.get_model("cms", "Title")

    for course_run in CourseRun.objects.iterator():
        try:
            course_run.direct_course = Course.objects.only("id").get(
                # Target courses that are ancestors of the course run
                extended_object__node=course_run.extended_object.node.parent,
                # Get the course in the same version as the course run
                extended_object__publisher_is_draft=course_run.extended_object.publisher_is_draft,
            )
        except Course.DoesNotExist as err:
            raise BadMigrationError(
                f"Course run {course_run.id:d} is not child of a course page."
            )

        # Create Django Parler translation objects for each language to hold the title
        for title_obj in Title.objects.filter(page=course_run.extended_object):
            CourseRunTranslation.objects.update_or_create(
                master=course_run,
                language_code=title_obj.language,
                defaults={"title": title_obj.title},
            )

        course_run.save()

    # Link draft and public instances with the new OneToOneField
    for course_run in CourseRun.objects.iterator():
        if course_run.public_extension_id:
            CourseRun.objects.filter(id=course_run.public_extension_id).update(
                draft_course_run=course_run
            )

    Page.objects.filter(courserun__isnull=False).delete()


class Migration(migrations.Migration):

    dependencies = [
        ("courses", "0021_auto_20201202_1146"),
    ]

    operations = [
        migrations.RunPython(populate_direct_course, migrations.RunPython.noop),
    ]



"""
Useful stuff for testing
"""
from cms.test_utils.testcases import CMSTestCase


class CMSPluginTestCase(CMSTestCase):
    """
    Enriched CMS test case object to include useful stuff about plugin
    rendering.

    Note:
        The CMSTestCase mixin from DjangoCMS is making queries in its "tearDown"
        method which causes database transaction errors with tests that have
        already generated a database error.

        A workaround is to decorate such tests with
        ``django.db.transaction`` (or to use it as a context manager) so that
        the test has its own transaction that is rolled back before the TearDown
        method sends its queries to the database.
    """

    def get_practical_plugin_context(self, extra_context=None):
        """
        Build a template context with dummy request object and
        instanciated content renderer suitable to perform full rendering of
        any plugin.

        Note:
            CMSTestCase use a dummy AnonymousUser on default behavior, you can
            override it with a custom user as an ``user`` attribute on your
            test case object. In most cases we should in fact define this
            attribute during test to use a UserFactory instead of a global
            user for every tests.

        Keyword Arguments:
            extra_context (dict): Dictionnary to add extra variable to context.
                Default to an empty dict.

        Returns:
            django.template.Context: Template context filled with request
            object as ``request`` item and content renderer as
            ``cms_content_renderer`` item.
        """
        context = self.get_context()
        if extra_context:
            context.update(extra_context)

        renderer = self.get_content_renderer(request=context["request"])

        # 'cms_content_renderer' is the attempted item name from CMS rendering
        # machinery
        context["cms_content_renderer"] = renderer

        return context



"""Error views"""
from django.shortcuts import render
from django.utils.translation import gettext_lazy as _

CONTEXT_ERRORS = {
    400: {
        "title": _("Bad request"),
        "content": _(
            "Your query seems to be malformed. Please check your data and try again."
        ),
    },
    403: {
        "title": _("Forbidden"),
        "content": _(
            "You do not have the required permissions to access this resource."
        ),
    },
    404: {
        "title": _("Page not found"),
        "content": _("The requested resource does not exist."),
    },
    500: {
        "title": _("Server bad request"),
        "content": _("Something went wrong. Please try again later."),
    },
}


def error_view_handler(request, exception, status):
    """
    Render error template with the related error context
    """
    if status not in [400, 403, 404, 500]:
        status = 500

    return render(
        request,
        template_name="richie/error.html",
        status=status,
        context={
            "error": exception,
            "status": status,
            "title": CONTEXT_ERRORS[status]["title"],
            "content": CONTEXT_ERRORS[status]["content"],
        },
    )


def error_400_view_handler(request, exception=None):
    """Error 400 view"""
    return error_view_handler(request, exception, 400)


def error_403_view_handler(request, exception=None):
    """Error 403 view"""
    return error_view_handler(request, exception, 403)


def error_404_view_handler(request, exception=None):
    """Error 404 view"""
    return error_view_handler(request, exception, 404)


def error_500_view_handler(request, exception=None):
    """Error 500 view"""
    return error_view_handler(request, exception, 500)



"""
Utils that can be useful throughout Richie's courses app
"""
import hashlib
import hmac

from django.utils.text import slugify


def normalize_code(code):
    """Normalize object codes to avoid duplicates."""
    return slugify(code, allow_unicode=True).upper() if code else None


def get_signature(message, secret):
    """Return a SHA256 signature for the message."""
    digest = hmac.new(
        secret.encode("utf-8"),
        msg=message.encode("utf-8"),
        digestmod=hashlib.sha256,
    ).hexdigest()
    return f"SIG-HMAC-SHA256 {digest:s}"



"""
CMS Wizard to add a course page
"""
from django import forms
from django.conf import settings
from django.core import validators
from django.core.exceptions import PermissionDenied
from django.template.defaultfilters import slugify
from django.utils.functional import cached_property
from django.utils.translation import get_language
from django.utils.translation import gettext_lazy as _

from cms.api import add_plugin, create_page
from cms.cms_wizards import (
    CMSPageWizard,
    CMSSubPageWizard,
    cms_page_wizard,
    cms_subpage_wizard,
)
from cms.forms.wizards import CreateCMSPageForm, CreateCMSSubPageForm, SlugWidget
from cms.models import Page
from cms.utils.page_permissions import user_can_change_page
from cms.wizards.forms import BaseFormMixin
from cms.wizards.wizard_base import Wizard
from cms.wizards.wizard_pool import wizard_pool

from . import defaults, models, utils


class ExcludeRootReverseIDMixin:
    """A mixin for wizards to disable it on portion of sites that are managed by Richie."""

    def user_has_add_permission(self, user, page):
        """Check that the page or any of its ancestors is not a special Richie page."""
        if (
            page.reverse_id in defaults.ROOT_REVERSE_IDS
            or page.get_ancestor_pages()
            .filter(reverse_id__in=defaults.ROOT_REVERSE_IDS)
            .exists()
        ):
            return False
        return super().user_has_add_permission(user, page)


class RichieCMSPageWizard(ExcludeRootReverseIDMixin, CMSPageWizard):
    """A page wizard that can not be created below our Richie extended pages."""


class RichieCMSSubPageWizard(ExcludeRootReverseIDMixin, CMSSubPageWizard):
    """A subpage wizard that can not be created below our Richie extended pages."""


wizard_pool.unregister(cms_page_wizard)
wizard_pool.register(
    RichieCMSPageWizard(
        title=_("New page"),
        weight=100,
        form=CreateCMSPageForm,
        model=Page,
        description=_("Create a new page next to the current page."),
    )
)

wizard_pool.unregister(cms_subpage_wizard)
wizard_pool.register(
    RichieCMSSubPageWizard(
        title=_("New sub page"),
        weight=110,
        form=CreateCMSSubPageForm,
        model=Page,
        description=_("Create a page below the current page."),
    )
)


class BaseWizardForm(BaseFormMixin, forms.Form):
    """
    Factorize the part of the wizard form that handles the title and the slug as they are
    common to all our wizard forms.
    """

    title = forms.CharField(
        max_length=255,
        required=True,
        widget=forms.TextInput(),
        label=_("Page title"),
        help_text=_("Title of the page in current language"),
    )
    slug = forms.CharField(
        max_length=200,  # Should be less than 255, the "max_length" of a Page's "path" field
        required=False,
        widget=SlugWidget(),
        validators=[validators.validate_slug],
        label=_("Page slug"),
        help_text=_("Slug of the page in current language"),
    )

    def clean(self):
        """
        Ensure that the slug field is set or derive it from the title
        """
        cleaned_data = super().clean()

        # If the slug is not explicitly set, generate it from the title
        if cleaned_data.get("title") and not cleaned_data.get("slug"):
            cleaned_data["slug"] = slugify(cleaned_data["title"])[:200]

        # Check that the length of the slug is compatible with its parent page:
        #  a page `path` is limited to 255 chars, therefore the course page slug should
        # always be shorter than (255 - length of parent page path - 1 character for the "/")
        length = len(self.parent_page.get_path()) + 1 + len(cleaned_data["slug"])
        if length > 255:
            raise forms.ValidationError(
                {
                    "slug": [
                        _(
                            "This slug is too long. The length of the path built by "
                            "prepending the slug of the parent page would be {:d} "
                            "characters long and it should be less than 255"
                        ).format(length)
                    ]
                }
            )

        if (
            self.parent_page.get_child_pages()
            .filter(title_set__slug=cleaned_data["slug"])
            .exists()
        ):
            raise forms.ValidationError({"slug": [_("This slug is already in use")]})

        return cleaned_data

    @cached_property
    def parent_page(self):
        """
        The parent page for each type of page is defined on its page extension model.
        """
        try:
            return Page.objects.get(
                reverse_id=self.model.PAGE["reverse_id"], publisher_is_draft=True
            )
        except Page.DoesNotExist as error:
            reverse_id = self.model.PAGE["reverse_id"]
            raise forms.ValidationError(
                {
                    "slug": [
                        _(
                            "You must first create a parent page and set its `reverse_id` to "
                            "`{:s}`."
                        ).format(reverse_id)
                    ]
                }
            ) from error

    def save(self):
        """
        Create the page with "title" and "slug"
        """
        return create_page(
            title=self.cleaned_data["title"],
            slug=self.cleaned_data["slug"],
            language=get_language(),
            parent=self.parent_page,
            template=self.model.PAGE["template"],
            published=False,  # The creation wizard should not publish the page
        )


class CourseWizardForm(BaseWizardForm):
    """
    This form is used by the wizard that creates a new course page.
    A related Course model is created for each course page.
    """

    code = forms.CharField(
        max_length=models.Course.CODE_MAX_LENGTH,
        required=getattr(settings, "RICHIE_COURSE_CODE_REQUIRED", False),
        label=_("Code"),
        help_text=_("Unique reference for the course."),
    )

    model = models.Course

    def clean(self):
        """
        Permission to add a course was already checked when we displayed the list of wizard
        entries, but we need to prevent form hacking.
        """
        # The current page should be an organization
        if not self.page:
            raise PermissionDenied()

        try:
            self.page.organization
        except models.Organization.DoesNotExist as error:
            raise PermissionDenied() from error

        # The user should be allowed to modify this organization page
        if not user_can_change_page(self.user, self.page):
            raise PermissionDenied()

        # The user should have permission to create a course object
        if not (self.user.is_staff and self.user.has_perm("courses.add_course")):
            raise PermissionDenied()

        return super().clean()

    def clean_code(self):
        """Ensure that the code field is unique among pages."""
        code = utils.normalize_code(self.cleaned_data["code"])
        if code and models.Course.objects.filter(code=code).exists():
            raise forms.ValidationError("A course already exists with this code.")

        return code

    def save(self):
        """
        The parent form created the page.
        This method creates the associated course page extension.
        If the current page is an organization, the new course should get attached to it via a
        plugin, and the admin group of the organization should get admin access to the course.
        """
        page = super().save()
        course = models.Course.objects.create(
            extended_object=page, code=self.cleaned_data["code"]
        )
        course.create_page_role()

        try:
            organization = self.page.organization
        except models.Organization.DoesNotExist:
            pass
        else:
            # Add a plugin for the organization
            placeholder = course.extended_object.placeholders.get(
                slot="course_organizations"
            )
            add_plugin(
                language=get_language(),
                placeholder=placeholder,
                plugin_type="OrganizationPlugin",
                **{"page": self.page},
            )
            course.create_permissions_for_organization(organization)

        return page


class CourseWizard(Wizard):
    """Inherit from Wizard because each wizard must have its own Python class."""

    def user_has_add_permission(self, user, **kwargs):
        """
        Returns: True if the user has the permission to add course objects, False otherwise.

        A course can only be created when visiting an organization page (that will become the
        course's main organization) that the user has the right to update.
        """
        if not kwargs.get("page"):
            return False

        try:
            organization = kwargs["page"].organization
        except models.Organization.DoesNotExist:
            return False

        return (
            user.has_perm("courses.add_course")
            and user_can_change_page(user, organization.extended_object)
            and super().user_has_add_permission(user, **kwargs)
        )


wizard_pool.register(
    CourseWizard(
        title=_("New course page"),
        description=_("Create a new course page"),
        model=models.Course,
        form=CourseWizardForm,
        weight=200,
    )
)


class OrganizationWizardForm(BaseWizardForm):
    """
    This form is used by the wizard that creates a new organization page
    A related organization model is created for each organization page
    """

    code = forms.CharField(
        max_length=models.Organization.CODE_MAX_LENGTH,
        required=getattr(settings, "RICHIE_ORGANIZATION_CODE_REQUIRED", False),
        label=_("Code"),
        help_text=_("Unique reference for the organization."),
    )

    model = models.Organization

    def clean(self):
        """
        Permission to add an organization was already checked when we displayed the list of
        wizard entries, but we need to prevent form hacking.
        """
        # The user should have permission to create an organization object
        if not (self.user.is_staff and self.user.has_perm("courses.add_organization")):
            raise PermissionDenied()

        return super().clean()

    def save(self):
        """
        The parent form created the page.
        This method creates the associated organization.
        """
        page = super().save()
        organization = models.Organization.objects.create(
            extended_object=page, code=self.cleaned_data["code"]
        )
        organization.create_page_role()
        return page

    def clean_code(self):
        """Ensure that the code field is unique among pages."""
        code = utils.normalize_code(self.cleaned_data["code"])
        if code and models.Organization.objects.filter(code=code).exists():
            raise forms.ValidationError(
                "An organization already exists with this code."
            )

        return code


class OrganizationWizard(Wizard):
    """Inherit from Wizard because each wizard must have its own Python class."""

    def user_has_add_permission(self, user, **kwargs):
        """
        Returns: True if the user has the permission to add organization objects, False otherwise.
        """
        return user.has_perm(
            "courses.add_organization"
        ) and super().user_has_add_permission(user, **kwargs)


wizard_pool.register(
    OrganizationWizard(
        title=_("New organization page"),
        description=_("Create a new organization page"),
        model=models.Organization,
        form=OrganizationWizardForm,
        weight=200,
    )
)


class CategoryWizardForm(BaseWizardForm):
    """
    This form is used by the wizard that creates a new category page.
    A related Category model is created for each category page.
    """

    model = models.Category

    def clean(self):
        """
        Permission to add an category was already checked when we displayed the list of
        wizard entries, but we need to prevent form hacking.
        """
        # The user should have permission to create a category object
        if not (self.user.is_staff and self.user.has_perm("courses.add_category")):
            raise PermissionDenied()

        return super().clean()

    @cached_property
    def parent_page(self):
        """
        If the current page is a category, the new category should be created as a child.
        Otherwise, it defaults to the categories root page defined by the `ROOT_REVERSE_ID`
        property of the Category model.
        """
        try:
            self.page.category
        except models.Category.DoesNotExist:
            return super().parent_page
        else:
            return self.page

    def save(self):
        """
        The parent form created the page.
        This method creates the associated category page extension.
        """
        page = super().save()
        models.Category.objects.create(extended_object=page)
        return page


class CategoryWizard(Wizard):
    """Inherit from Wizard because each wizard must have its own Python class."""

    def user_has_add_permission(self, user, **kwargs):
        """
        Returns: True if the user has the permission to add category objects, False otherwise.
        """
        return user.has_perm(
            "courses.add_category"
        ) and super().user_has_add_permission(user, **kwargs)


wizard_pool.register(
    CategoryWizard(
        title=_("New category page"),
        description=_("Create a new category page"),
        model=models.Category,
        form=CategoryWizardForm,
        weight=200,
    )
)


class BlogPostWizardForm(BaseWizardForm):
    """
    This form is used by the wizard that creates a new blog post
    A related blogpost model is created for each blog post
    """

    model = models.BlogPost

    def clean(self):
        """
        Permission to add an blog post was already checked when we displayed the list of
        wizard entries, but we need to prevent form hacking.
        """
        # The user should have permission to create a blog post object
        if not (self.user.is_staff and self.user.has_perm("courses.add_blogpost")):
            raise PermissionDenied()

        return super().clean()

    def save(self):
        """
        The parent form created the page.
        This method creates the associated blogpost.
        """
        page = super().save()
        models.BlogPost.objects.create(extended_object=page)
        return page


class BlogPostWizard(Wizard):
    """Inherit from Wizard because each wizard must have its own Python class."""

    def user_has_add_permission(self, user, **kwargs):
        """
        Returns: True if the user has the permission to add blog post objects, False otherwise.
        """
        return user.has_perm(
            "courses.add_blogpost"
        ) and super().user_has_add_permission(user, **kwargs)


wizard_pool.register(
    BlogPostWizard(
        title=_("New blog post"),
        description=_("Create a new blog post"),
        model=models.BlogPost,
        form=BlogPostWizardForm,
        weight=200,
    )
)


class PersonWizardForm(BaseWizardForm):
    """
    This form is used by the wizard that creates a new person page.
    A related Person model is created for each person page.
    """

    model = models.Person

    def clean(self):
        """
        Permission to add an person was already checked when we displayed the list of
        wizard entries, but we need to prevent form hacking.
        """
        # The user should have permission to create a person object
        if not (self.user.is_staff and self.user.has_perm("courses.add_person")):
            raise PermissionDenied()

        return super().clean()

    def save(self):
        """
        The parent form created the page.
        This method creates the associated person page extension.
        """
        page = super().save()
        models.Person.objects.create(extended_object=page)
        return page


class PersonWizard(Wizard):
    """Inherit from Wizard because each wizard must have its own Python class."""

    def user_has_add_permission(self, user, **kwargs):
        """
        Returns: True if the user has the permission to add person objects, False otherwise.
        """
        return user.has_perm("courses.add_person") and super().user_has_add_permission(
            user, **kwargs
        )


wizard_pool.register(
    PersonWizard(
        title=_("New person page"),
        description=_("Create a new person page"),
        model=models.Person,
        form=PersonWizardForm,
        weight=200,
    )
)


class ProgramWizardForm(BaseWizardForm):
    """
    This form is used by the wizard that creates a new program
    A related program model is created for each program
    """

    model = models.Program

    def clean(self):
        """
        Permission to add a program was already checked when we displayed the list of
        wizard entries, but we need to prevent form hacking.
        """
        # The user should have permission to create a program object
        if not (self.user.is_staff and self.user.has_perm("courses.add_program")):
            raise PermissionDenied()

        return super().clean()

    def save(self):
        """
        The parent form created the page.
        This method creates the associated program.
        """
        page = super().save()
        models.Program.objects.create(extended_object=page)
        return page


class ProgramWizard(Wizard):
    """Inherit from Wizard because each wizard must have its own Python class."""

    def user_has_add_permission(self, user, **kwargs):
        """
        Returns: True if the user has the permission to add program objects, False otherwise.
        """
        return user.has_perm("courses.add_program") and super().user_has_add_permission(
            user, **kwargs
        )


wizard_pool.register(
    ProgramWizard(
        title=_("New program"),
        description=_("Create a new program"),
        model=models.Program,
        form=ProgramWizardForm,
        weight=200,
    )
)



"""Update Elasticsearch indices each time a page is modified."""
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction

from cms.models import Title

from richie.apps.courses.models import Category, Course, Organization, Person
from richie.apps.search.index_manager import richie_bulk
from richie.apps.search.indexers import ES_INDICES


def apply_es_action_to_course(instance, action, _language):
    """
    Update Elasticsearch indices when a course is modified:
    - update the course document in the Elasticsearch courses index.

    Returns None if the page was related to a course and the Elasticsearch update is done.
    Raises ObjectDoesNotExist if the page instance is not related to a course.
    """
    course = Course.objects.get(draft_extension__extended_object=instance)
    if not course.is_snapshot:
        richie_bulk(
            [ES_INDICES.courses.get_es_document_for_course(course, action=action)]
        )


def apply_es_action_to_organization(instance, action, language):
    """
    Update Elasticsearch indices when an organization is modified:
    - update the organization document in the Elasticsearch organizations index for the
      organization and its direct parent (because the parent ID may change from Parent to Leaf),
    - update the course documents in the Elasticsearch courses index for all courses linked to
      this organization.

    Returns None if the page was related to an organization and the Elasticsearch update is done.
    Raises ObjectDoesNotExist if the page instance is not related to an organization.
    """
    organization = Organization.objects.get(draft_extension__extended_object=instance)
    actions = [
        ES_INDICES.courses.get_es_document_for_course(course)
        for course in organization.get_courses(language)
        if not course.is_snapshot
    ]
    actions.append(
        ES_INDICES.organizations.get_es_document_for_organization(
            organization, action=action
        )
    )

    # Update the organization's parent only if it exists
    try:
        parent = organization.extended_object.get_parent_page().organization
    except AttributeError:
        pass
    else:
        actions.append(
            ES_INDICES.organizations.get_es_document_for_organization(parent)
        )

    richie_bulk(actions)


def apply_es_action_to_person(instance, action, language):
    """
    Update Elasticsearch indices when a person is modified:
    - update the person document in the Elasticsearch persons index for the
      person,
    - update the course documents in the Elasticsearch courses index for all courses linked to
      this person.

    Returns None if the page was related to a person and the Elasticsearch update is done.
    Raises ObjectDoesNotExist if the page instance is not related to a person.
    """
    person = Person.objects.get(draft_extension__extended_object=instance)
    actions = [
        ES_INDICES.courses.get_es_document_for_course(course)
        for course in person.get_courses(language)
        if not course.is_snapshot
    ]
    actions.append(ES_INDICES.persons.get_es_document_for_person(person, action=action))

    richie_bulk(actions)


def apply_es_action_to_category(instance, action, language):
    """
    Update Elasticsearch indices when a category is modified:
    - update the category document in the Elasticsearch categories index for the category
      and its direct parent (because the parent ID may change from Parent to Leaf),
    - update the course documents in the Elasticsearch courses index for all courses linked to
      this category.

    Returns None if the page was related to a category and the Elasticsearch update is done.
    Raises ObjectDoesNotExist if the page instance is not related to a category.
    """
    category = Category.objects.get(draft_extension__extended_object=instance)
    actions = [
        ES_INDICES.courses.get_es_document_for_course(course)
        for course in category.get_courses(language)
        if not course.is_snapshot
    ]
    actions.append(
        ES_INDICES.categories.get_es_document_for_category(category, action=action)
    )

    # Update the category's parent only if it exists
    try:
        parent = category.extended_object.get_parent_page().category
    except AttributeError:
        pass
    else:
        actions.append(ES_INDICES.categories.get_es_document_for_category(parent))

    richie_bulk(actions)


def apply_es_action_to_page(page, action, language):
    """
    Try updating each type of page extension one-by-one until one works (because we don't know
    to which type of page extension this page is related).
    """
    for method in [
        apply_es_action_to_course,
        apply_es_action_to_category,
        apply_es_action_to_organization,
        apply_es_action_to_person,
    ]:
        try:
            # The method should raise an ObjectDoesNotExist exception if the page extension
            # linked to this page is of another type.
            method(page, action, language)
        except ObjectDoesNotExist:
            continue
        else:
            return


# pylint: disable=unused-argument
def on_page_published(sender, instance, language, **kwargs):
    """
    Trigger update of the Elasticsearch indices impacted by the modification of the instance
    only once the database transaction is successful.
    """
    if getattr(settings, "RICHIE_KEEP_SEARCH_UPDATED", True):
        transaction.on_commit(
            lambda: apply_es_action_to_page(instance, "index", language)
        )


# pylint: disable=unused-argument
def on_page_unpublished(sender, instance, language, **kwargs):
    """
    Trigger update of the Elasticsearch indices impacted by the modification of the instance
    only once the database transaction is successful.
    """
    if getattr(settings, "RICHIE_KEEP_SEARCH_UPDATED", True):
        # Only unlist pages that are unpublished from all languages otherwise,
        # reindex it to remove the unpublished language
        action = (
            "index"
            if Title.objects.filter(page=instance, published=True).exists()
            else "delete"
        )
        transaction.on_commit(
            lambda: apply_es_action_to_page(instance, action, language)
        )



"""
Declare and configure the models for the blog part
"""
from django.db import models
from django.utils import translation
from django.utils.translation import gettext_lazy as _

from cms.api import Page
from cms.extensions.extension_pool import extension_pool
from cms.models.pluginmodel import CMSPlugin

from ...core.models import BasePageExtension, get_plugin_language_fallback_clause
from .. import defaults
from .category import Category, CategoryPluginModel


class BlogPost(BasePageExtension):
    """
    The blogpost extension represents and records a blog article.
    """

    PAGE = defaults.BLOGPOSTS_PAGE

    class Meta:
        db_table = "richie_blog_post"
        ordering = ["-pk"]
        verbose_name = _("blog post")
        verbose_name_plural = _("blog posts")

    def __str__(self):
        """Human representation of a blogpost"""
        model = self._meta.verbose_name.title()
        name = self.extended_object.get_title()
        return f"{model:s}: {name:s}"

    def get_categories(self, language=None):
        """
        Return the categories linked to the blog post via a category plugin in any of the
        placeholders on the blog post detail page, ranked by their `path` to respect the
        order in the categories tree.
        """
        return self.get_direct_related_page_extensions(
            Category, CategoryPluginModel, language=language
        )

    def get_related_blogposts(self, language=None):
        """
        Return the blogposts related to the current blogpost ie that share common categories.
        the blogposts returned are sorted by their `path` to respect the order in the page
        tree.
        """
        is_draft = self.extended_object.publisher_is_draft
        current_language = language or translation.get_language()
        language_clause = get_plugin_language_fallback_clause(
            current_language, is_draft
        )

        bfs = (
            "extended_object__placeholders__cmsplugin__courses_categorypluginmodel"
            "__page__category__in"
        )
        selector = {bfs: self.get_categories()}

        # pylint: disable=no-member
        return (
            BlogPost.objects.filter(
                language_clause,
                extended_object__publisher_is_draft=is_draft,
                **selector,
            )
            .exclude(id=self.id)
            .select_related("extended_object")
            .distinct()
            .order_by("-extended_object__publication_date")
        )


class BlogPostPluginModel(CMSPlugin):
    """
    BlogPost plugin model handles the relation between the BlogPostPluginModel
    and its blogpost instance
    """

    page = models.ForeignKey(
        Page,
        on_delete=models.CASCADE,
        related_name="blogpost_plugins",
        limit_choices_to={"publisher_is_draft": True, "blogpost__isnull": False},
    )
    variant = models.CharField(
        _("variant"),
        max_length=50,
        choices=defaults.BLOGPOST_GLIMPSE_VARIANT_CHOICES,
        help_text=_("Optional glimpse variant for a custom look."),
        blank=True,
        null=True,
    )

    class Meta:
        db_table = "richie_blog_post_plugin"
        verbose_name = _("blog post plugin")
        verbose_name_plural = _("blog post plugins")

    def __str__(self):
        """Human representation of a page plugin"""
        return self.page.get_title()


extension_pool.register(BlogPost)



"""Signals to update the Elasticsearch indices when page modifications are published."""
from django.apps import AppConfig


class SearchConfig(AppConfig):
    """Configuration class for the search app."""

    name = "richie.apps.search"
    verbose_name = "Richie search app"

    # pylint: disable=import-outside-toplevel
    def ready(self):
        """Register signals to update the Elasticsearch indices."""
        from cms.signals import post_publish, post_unpublish

        from .signals import on_page_published, on_page_unpublished

        post_publish.connect(on_page_published, dispatch_uid="search_post_publish")
        post_unpublish.connect(
            on_page_unpublished, dispatch_uid="search_post_unpublish"
        )



# Generated by Django 3.1.3 on 2020-11-18 10:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("courses", "0019_auto_20201117_1004"),
    ]

    operations = [
        migrations.AlterField(
            model_name="categorypluginmodel",
            name="variant",
            field=models.CharField(
                blank=True,
                choices=[
                    (None, "Inherit"),
                    ("glimpse", "Default"),
                    ("badge", "Badge"),
                    ("tag", "Tag"),
                ],
                help_text="Optional glimpse variant for a custom look.",
                max_length=50,
                null=True,
                verbose_name="variant",
            ),
        ),
    ]



"""
LTIConsumer CMS plugin factories
"""
from django.conf import settings

import factory
import factory.fuzzy

from .models import LTIConsumer


class LTIConsumerFactory(factory.django.DjangoModelFactory):
    """
    Factory to create random instances of LTIConsumer for testing.
    """

    class Meta:
        model = LTIConsumer

    lti_provider_id = factory.fuzzy.FuzzyChoice(
        getattr(settings, "RICHIE_LTI_PROVIDERS", {}).keys()
    )
    is_automatic_resizing = factory.fuzzy.FuzzyChoice([True, False])
    inline_ratio = factory.fuzzy.FuzzyFloat(0.1, 10)

    @factory.lazy_attribute
    def url(self):
        """Generates a random url in accordance with the LTI provider."""
        if self.lti_provider_id:
            # Let the "save" method generate by returning None
            return None

        return factory.Faker("url").evaluate(
            None, None, {"locale": settings.LANGUAGE_CODE}
        )



"""LTI Consumer plugin URLs configuration."""

from django.urls import include, path

from rest_framework.routers import DefaultRouter

from . import models
from .api import LTIConsumerViewsSet

router = DefaultRouter()
router.register(
    models.LTIConsumer.RESOURCE_NAME,
    LTIConsumerViewsSet,
    basename="lti-consumer",
)

url_patterns = [path("", include(router.urls))]



"""
ElasticSearch person document management utilities
"""
from collections import defaultdict

from django.conf import settings
from django.utils import translation

from cms.models import Title
from djangocms_picture.models import Picture

from richie.plugins.simple_picture.helpers import get_picture_info
from richie.plugins.simple_text_ckeditor.models import SimpleText

from ...courses.models import Person
from ..defaults import ES_INDICES_PREFIX
from ..forms import ItemSearchForm
from ..text_indexing import MULTILINGUAL_TEXT
from ..utils.i18n import get_best_field_language
from ..utils.indexers import slice_string_for_completion


class PersonsIndexer:
    """
    Makes available the parameters the indexer requires as well as a function to shape
    objects into what we want to index in ElasticSearch
    """

    index_name = f"{ES_INDICES_PREFIX}_persons"
    form = ItemSearchForm
    mapping = {
        "dynamic_templates": MULTILINGUAL_TEXT,
        "properties": {
            # Searchable
            **{
                f"complete.{lang:s}": {
                    "type": "completion",
                    "analyzer": "simple_diacritics_insensitive",
                }
                for lang, _ in settings.LANGUAGES
            },
            **{
                "{key}.{lang}": {"type": "text", "analyzer": "simple"}
                for lang, _ in settings.LANGUAGES
                for key in ["bio", "title"]
            },
            # Not searchable
            "absolute_url": {"type": "object", "enabled": False},
            "portrait": {"type": "object", "enabled": False},
            # Create a raw title field to enable alphabetical sorting
            # We cannot use the default title field as the analysis prevents sorting on it
            **{
                f"title_raw.{lang}": {"type": "keyword"}
                for lang, _ in settings.LANGUAGES
            },
        },
    }
    scripts = {}
    display_fields = ["absolute_url", "portrait", "title"]

    @classmethod
    def get_es_document_for_person(cls, person, index=None, action="index"):
        """Build an Elasticsearch document from the person instance."""
        index = index or cls.index_name

        # Get published titles
        titles = {
            t.language: t.title
            for t in Title.objects.filter(page=person.extended_object, published=True)
        }

        # Prepare portrait images
        portrait_images = {}
        for portrait in Picture.objects.filter(
            cmsplugin_ptr__placeholder__page=person.extended_object,
            cmsplugin_ptr__placeholder__slot="portrait",
        ):
            language = portrait.cmsplugin_ptr.language
            with translation.override(language):
                portrait_images[language] = get_picture_info(portrait, "portrait")

        # Get bio texts
        bio = defaultdict(list)
        for simple_text in SimpleText.objects.filter(
            cmsplugin_ptr__placeholder__page=person.extended_object,
            cmsplugin_ptr__placeholder__slot="bio",
        ):
            bio[simple_text.cmsplugin_ptr.language].append(simple_text.body)

        return {
            "_id": str(person.extended_object_id),
            "_index": index,
            "_op_type": action,
            "absolute_url": {
                lang: person.extended_object.get_absolute_url(lang)
                for lang, _ in settings.LANGUAGES
            },
            "bio": {language: " ".join(st) for language, st in bio.items()},
            "complete": {
                language: slice_string_for_completion(title)
                for language, title in titles.items()
            },
            "portrait": portrait_images,
            "title": titles,
            "title_raw": titles,
        }

    @classmethod
    def get_es_documents(cls, index=None, action="index"):
        """
        Loop on all the persons in database and format them for the ElasticSearch index
        """
        index = index or cls.index_name

        for person in (
            Person.objects.filter(
                extended_object__publisher_is_draft=False,
                extended_object__title_set__published=True,
            )
            .distinct()
            .iterator()
        ):
            yield cls.get_es_document_for_person(person, index=index, action=action)

    @staticmethod
    def format_es_object_for_api(es_person, best_language):
        """
        Format an person stored in ES into a consistent and easy-to-consume record for
        API consumers
        """
        return {
            "id": es_person["_id"],
            "portrait": get_best_field_language(
                es_person["_source"]["portrait"], best_language
            ),
            "title": get_best_field_language(
                es_person["_source"]["title"], best_language
            ),
        }

    @staticmethod
    def format_es_document_for_autocomplete(es_document, language=None):
        """
        Format a document stored in ES into an easy-to-consume record for autocomplete consumers.
        This method differs from the regular one as objects retrieved from query VS complete
        queries can be formatted differently; and consumers of autocomplete do not need
        full objects.
        """
        return {
            "id": es_document["_id"],
            "kind": "persons",
            "title": get_best_field_language(es_document["_source"]["title"], language),
        }



# Generated by Django 3.1.7 on 2021-03-02 12:57

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("cms", "0022_auto_20180620_1551"),
    ]

    operations = [
        migrations.CreateModel(
            name="LTIConsumer",
            fields=[
                (
                    "cmsplugin_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="lti_consumer_lticonsumer",
                        serialize=False,
                        to="cms.cmsplugin",
                    ),
                ),
                (
                    "url",
                    models.URLField(
                        blank=True,
                        help_text="For a predefined provider, leave this field empty for uploading new content.",
                        verbose_name="LTI url",
                    ),
                ),
                (
                    "lti_provider_id",
                    models.CharField(
                        blank=True,
                        choices=[("lti_provider_test", "LTI Provider Test Video")],
                        help_text="Please choose a predefined provider or fill fields below.",
                        max_length=50,
                        null=True,
                        verbose_name="Predefined LTI provider",
                    ),
                ),
                (
                    "oauth_consumer_key",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
                (
                    "shared_secret",
                    models.CharField(blank=True, max_length=50, null=True),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("cms.cmsplugin",),
        ),
    ]



# Generated by Django 3.1.3 on 2020-11-23 10:46

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("courses", "0022_auto_20201202_convert_course_runs_to_simple_models"),
    ]

    operations = [
        migrations.AlterField(
            model_name="courserun",
            name="direct_course",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="runs",
                to="courses.course",
            ),
        ),
        migrations.RemoveField(
            model_name="courserun",
            name="extended_object",
        ),
        migrations.RemoveField(
            model_name="courserun",
            name="public_extension",
        ),
    ]



"""Default settings for Richie's LTI consumer plugin."""
from django.conf import settings
from django.utils.translation import gettext_lazy as _

STUDENT = "student"
INSTRUCTOR = "instructor"
PREDEFINED_LTI_PROVIDERS = [(None, _("Custom provider configuration"))] + [
    (provider_key, provider.get("display_name", provider_key))
    for provider_key, provider in getattr(settings, "RICHIE_LTI_PROVIDERS", {}).items()
]



# Generated by Django 2.2.11 on 2020-03-10 21:26

from django.db import migrations

from ..defaults import SECTION_TEMPLATES


def forwards_func(apps, schema_editor):
    """
    Update every section object which used the deprecated "section_cadenced"
    template to default section template (the first item from choices).
    """
    Section = apps.get_model("section", "Section")

    Section.objects.filter(template="richie/section/section_cadenced.html").update(
        template=SECTION_TEMPLATES[0][0]
    )


class Migration(migrations.Migration):

    dependencies = [
        ("section", "0003_auto_20191119_1650"),
    ]

    operations = [migrations.RunPython(forwards_func, migrations.RunPython.noop)]



"""
Section plugin forms
"""
from django import forms

from djangocms_text_ckeditor.widgets import TextEditorWidget

from .models import Section

CKEDITOR_CONFIGURATION_NAME = "CKEDITOR_INLINE_BOLD_CONFIGURATION"


class SectionForm(forms.ModelForm):
    """
    Plugin form used to fill its content from frontend admin.
    """

    class Meta:
        """
        Form meta attributes
        """

        model = Section
        widgets = {"title": TextEditorWidget(configuration=CKEDITOR_CONFIGURATION_NAME)}
        fields = {
            "title",
            "template",
        }



"""LTI Consumer plugin forms."""
import re

from django import forms
from django.conf import settings
from django.utils.translation import gettext_lazy as _

import exrex

from .models import LTIConsumer

SHARED_SECRET_PLACEHOLDER = "%%shared_secret_placeholder%%"  # nosec


class LTIConsumerForm(forms.ModelForm):
    """Plugin form used to add a LTI provided content"""

    form_shared_secret = forms.CharField(
        label=_("shared secret"),
        widget=forms.PasswordInput(
            render_value=True, attrs={"onfocus": "this.value=''"}
        ),
        required=False,
        max_length=50,
    )

    class Meta:
        """Form meta attributes"""

        model = LTIConsumer
        fields = [
            "lti_provider_id",
            "url",
            "oauth_consumer_key",
            "form_shared_secret",
            "is_automatic_resizing",
            "inline_ratio",
        ]

    def __init__(self, *args, **kwargs):
        """Initialize the "form_shared_secret" field with a placeholder value (whatever
        since it will be obfuscated... it is just here to show to the user that a shared
        secret is set).
        """
        super().__init__(*args, **kwargs)
        if self.instance and self.instance.shared_secret:
            self.fields["form_shared_secret"].initial = SHARED_SECRET_PLACEHOLDER

    def clean(self):
        """
        This form can be filled in two ways:
        - use a predefined LTI provider by filling lti_provider_id field
        - use a custom LTI provider by filling others fields

        Here we're adding related errors
        The shared secret field is de-synchronized from its corresponding field on the model:
        - a "form_shared_secret" field serves as interface in the UI
        - the field is of "password" type: its value is obfuscated
        - if "shared_secret" is filled, "form_shared_secret" is filled with a dummy value
        - if "form_shared_secret" is submitted with a value other than the dummy value,
          the "shared_secret" field is updated on the model.
        """
        provider_id = self.cleaned_data.get("lti_provider_id")
        shared_secret = self.cleaned_data.get("form_shared_secret")

        url = self.cleaned_data.get("url")

        if provider_id:
            if url:
                provider = getattr(settings, "RICHIE_LTI_PROVIDERS", {}).get(
                    provider_id, {}
                )
                is_regex = provider.get("is_base_url_regex", True)
                base_url = provider.get("base_url", "")

                if is_regex and not re.compile(base_url).search(url):
                    message = _(
                        'The url is not valid for this provider. It should be of the form "{:s}".'
                    ).format(exrex.getone(base_url))
                    self.add_error("url", message)

                if not is_regex and base_url not in url:
                    message = _(
                        'The url is not valid for this provider. It should start with "{:s}".'
                    ).format(base_url)
                    self.add_error("url", message)

            # Reset credentials
            self.cleaned_data["oauth_consumer_key"] = None
            self.instance.shared_secret = None
        else:
            oauth_consumer_key = self.cleaned_data.get("oauth_consumer_key")

            if not url and not oauth_consumer_key and not shared_secret:
                self.add_error(
                    "lti_provider_id",
                    _("Please choose a predefined provider, or fill fields below"),
                )
                message = _(
                    "Please choose a predefined provider above, or fill this field"
                )
                self.add_error("url", message)
                self.add_error("oauth_consumer_key", message)
                self.add_error("form_shared_secret", message)
            else:
                message = _("Please fill this field")
                if not url:
                    self.add_error("url", message)
                if not oauth_consumer_key:
                    self.add_error("oauth_consumer_key", message)
                if not shared_secret:
                    self.add_error("form_shared_secret", message)

        if shared_secret and shared_secret != SHARED_SECRET_PLACEHOLDER:
            self.instance.shared_secret = shared_secret

        return super().clean()



"""
API endpoints to access categories through ElasticSearch
"""
from django.conf import settings
from django.utils.translation import get_language_from_request

from elasticsearch.exceptions import NotFoundError
from rest_framework.decorators import action
from rest_framework.exceptions import NotFound
from rest_framework.response import Response
from rest_framework.viewsets import ViewSet

from .. import ES_CLIENT
from ..defaults import ES_PAGE_SIZE
from ..indexers import ES_INDICES
from ..utils.viewsets import ViewSetMetadata


class CategoriesViewSet(ViewSet):
    """
    A simple viewset with GET endpoints to fetch categories
    See API Blueprint for details on consumer use.
    """

    _meta = ViewSetMetadata(indexer=ES_INDICES.categories)

    # pylint: disable=no-self-use,unused-argument
    def list(self, request, version, kind):
        """
        Category search endpoint: pass query params to ElasticSearch so it filters categories
        and returns a list of matching items
        """
        # Instantiate the form to allow validation/cleaning
        params_form = self._meta.indexer.form(data=request.query_params)

        # Return a 400 with error information if the query params are not valid
        if not params_form.is_valid():
            return Response(status=400, data={"errors": params_form.errors})

        limit, offset, query = params_form.build_es_query(kind=kind)

        query["sort"] = [
            {f"title_raw.{get_language_from_request(request)}": {"order": "asc"}}
        ]

        try:
            # pylint: disable=unexpected-keyword-arg
            query_response = ES_CLIENT.search(
                _source=getattr(self._meta.indexer, "display_fields", "*"),
                index=self._meta.indexer.index_name,
                body=query,
                # Directly pass meta-params through as arguments to the ES client
                from_=offset,
                size=limit or getattr(settings, "RICHIE_ES_PAGE_SIZE", ES_PAGE_SIZE),
            )
        except NotFoundError as error:
            raise NotFound from error

        # Format the response in a consumer-friendly way
        # NB: if there are 0 hits the query_response is formatted the exact same way, only the
        # .hits.hits array is empty.
        response_object = {
            "meta": {
                "count": len(query_response["hits"]["hits"]),
                "offset": offset,
                "total_count": query_response["hits"]["total"]["value"],
            },
            "objects": [
                self._meta.indexer.format_es_object_for_api(
                    category,
                    # Get the best language we can return multilingual fields in
                    get_language_from_request(request),
                )
                for category in query_response["hits"]["hits"]
            ],
        }

        return Response(response_object)

    # pylint: disable=no-self-use,invalid-name,unused-argument
    def retrieve(self, request, pk, version, kind):
        """
        Return a single item by ID
        """
        # Wrap the ES get in a try/catch to we control the exception we emit — it would
        # raise and end up in a 500 error otherwise
        try:
            query_response = ES_CLIENT.get(
                index=self._meta.indexer.index_name,
                id=pk,
            )
        except NotFoundError as error:
            raise NotFound from error

        # Format a clean category object as a response
        return Response(
            self._meta.indexer.format_es_object_for_api(
                query_response,
                # Get the best language we can return multilingual fields in
                get_language_from_request(request),
            )
        )

    @action(detail=False)
    def autocomplete(self, request, version, kind):
        """
        Use the "completion" field on the categories mapping & objects to provide autocomplete
        functionality through an API endpoint.
        We cannot reuse the `AutocompleteMixin` as it does not handle "kinds" as categories need.
        """
        # Get a hold of the relevant indexer
        indexer = self._meta.indexer

        # Query our specific ES completion field
        language = get_language_from_request(request)
        autocomplete_query_response = ES_CLIENT.search(
            index=indexer.index_name,
            body={
                "suggest": {
                    "categories": {
                        "prefix": request.query_params["query"],
                        "completion": {
                            "contexts": {"kind": [kind]},
                            "field": f"complete.{language:s}",
                        },
                    }
                }
            },
        )

        # Build a response array from the list of completion options
        return Response(
            [
                indexer.format_es_document_for_autocomplete(
                    option, get_language_from_request(request)
                )
                for option in autocomplete_query_response["suggest"]["categories"][0][
                    "options"
                ]
            ]
        )



"""
Plugins urls
"""

from django.urls import include, path

from richie.plugins.lti_consumer.urls import url_patterns as lti_consumer_url_patterns

urlpatterns = [path("plugins/", include([*lti_consumer_url_patterns]))]



"""
i18n utilities for our search app
"""
from django.conf import settings


def get_best_field_language(multilingual_field, best_language):
    """
    Pick the best available language from a multilingual field.

    A multilingual field is eg:
        'title': {
            'es': 'mi título',
            'fr': 'mon titre',
        }

    1. Use the most appropriate language as determined by the consumer
    2. Default to language #0, then #1, then #2, etc. in settings.LANGUAGES
    """
    for language in [best_language] + [lang for lang, _ in settings.LANGUAGES]:
        try:
            return multilingual_field[language]
        except KeyError:
            pass

    return None



"""Make all filter definitions available from richie.apps.search.filter_definitions."""
from django.conf import settings
from django.utils.module_loading import import_string

# pylint: disable=unused-import
from ..defaults import FILTERS_CONFIGURATION
from .base import NestingWrapper  # noqa
from .courses import (  # noqa
    AvailabilityFilterDefinition,
    IndexableFilterDefinition,
    IndexableMPTTFilterDefinition,
    LanguagesFilterDefinition,
    StaticChoicesFilterDefinition,
)

FILTERS = {
    params["name"]: import_string(path)(**params)
    for path, params in getattr(
        settings, "RICHIE_FILTERS_CONFIGURATION", FILTERS_CONFIGURATION
    )
}



"""
Add full URL to a static Django file.
https://stackoverflow.com/a/47336360
"""

from django import template
from django.templatetags import static

# pylint: disable=invalid-name
register = template.Library()


class FullStaticNode(static.StaticNode):
    """
    Generate absolute full URL to a static file.
    """

    def url(self, context):
        request = context["request"]
        return request.build_absolute_uri(super().url(context))


@register.tag("static_absolute")
def do_static(parser, token):
    """
    Register the tag static_absolute
    """
    return FullStaticNode.handle_token(parser, token)



"""Define FilterDefinition classes for the courses index."""
import re
from operator import itemgetter

from django import forms
from django.core.exceptions import ImproperlyConfigured
from django.utils import timezone, translation
from django.utils.translation import gettext as _

from cms.api import Page

from richie.apps.core.defaults import ALL_LANGUAGES_DICT

from .. import ES_CLIENT
from ..fields.array import ArrayField
from ..indexers import ES_INDICES
from ..utils.i18n import get_best_field_language
from .base import BaseChoicesFilterDefinition, BaseFilterDefinition
from .helpers import applicable_facet_limit
from .mixins import (
    ChoicesAggsMixin,
    ChoicesQueryMixin,
    NestedChoicesAggsMixin,
    TermsQueryMixin,
)


class IndexableFilterDefinition(TermsQueryMixin, BaseFilterDefinition):
    """
    Filter definition for a terms-based filter. The choices are generated dynamically from
    the incoming facets to avoid having to hold in memory or iterate over an unbounded
    number of choices.
    """

    def __init__(self, name, reverse_id=None, **kwargs):
        self.reverse_id = reverse_id
        self._base_page = None
        super().__init__(name, **kwargs)

    @property
    def base_page(self):
        """
        Returns the base page under which the indexable pages target by this filter are placed
        in the CMS.

        The page is cached the first time it is requested and remains in cache as long as the
        application is running.
        """
        if self.reverse_id:
            if self._base_page is None:
                try:
                    page = Page.objects.select_related("node").get(
                        publisher_is_draft=False, reverse_id=self.reverse_id
                    )
                except Page.DoesNotExist:
                    return None
                else:
                    self._base_page = page

            return self._base_page

        return None

    @property
    def aggs_include(self):
        """Do not limit which facets are computed by default."""
        return ".*"

    # pylint: disable=unused-argument,arguments-differ
    def get_aggs_fragment(self, queries, data, *args, **kwargs):
        """
        Build the aggregations as a term query that counts all the different values assigned
        to the field.
        """
        # Look for an include parameter in the form data and default to the regex returned
        # by the `aggs_include` property:
        include = data[f"{self.name:s}_include"] or self.aggs_include

        # Use all the query fragments from the queries *but* the one(s) that filter on the
        # current filter
        filter_fragments = [
            clause
            for kf_pair in queries
            for clause in kf_pair["fragment"]
            if kf_pair["key"] is not self.name
        ]

        # Detect the applicable facet counts limit depending on the request
        base_facet_limit = applicable_facet_limit(data, self.name)

        # Add a control length to let us know if there are more items than what what we're
        # returning. This is achieved by counting the results in `get_faceted_definitions`.
        # Example:
        #   - applicable limit is 10;
        #   - 2 active filters;
        #   - query the top 13 (10 + 2 + 1) facets;
        #   => we'll be able to determine if there are more items depending on what is returned
        #      and how they are ordered. (see in `get_faceted_definitions`).
        control_length = len(data[self.name]) + 1

        # Terms aggregation will return the n top facet counts among all the values taken
        # by this field
        terms_aggs = {
            self.name: {
                # Rely on the built-in "terms" aggregation to get everything we need
                "aggregations": {
                    self.name: {
                        "terms": {
                            "field": self.term,
                            "include": include,
                            # Always force `min_doc_count` at 0: we filter the results in
                            # `get_faceted_definitions` so we can have more information on the
                            # availability of more values below the filter's `min_doc_count`.
                            "min_doc_count": 0,
                            "size": base_facet_limit + control_length,
                        }
                    }
                },
                # Use all the query fragments from the queries *but* the one(s) that
                # filter on the current filter, as it is handled by ElasticSearch for us
                "filter": {"bool": {"must": filter_fragments}},
            }
        }

        # Filters aggregation for values that were selected in the querystring (we must force
        # them because they may not be in the n top facet counts but we must make sure we keep
        # them so that they remain available as options that the user can see and unselect)
        terms_aggs.update(
            {
                # Create a custom aggregation for each value selected in the querystring
                # and matching the include regex
                # eg `organizations@P-0001` & `organizations@P-0002`
                f"{self.name:s}@{value:s}": {
                    "filter": {
                        "bool": {
                            # Use all the query fragments from the queries *but* the one(s) that
                            # filter on the current filter: we manually add back the only one that
                            # is relevant to the current choice.
                            "must": [{"term": {self.term: value}}]
                            + filter_fragments
                        }
                    }
                }
                for value in data.get(self.name, [])
                # The Elasticsearch include regex matches exact values so we must do the same
                # by adding ^ (resp. $) at the beginning (resp. at the end) of the pattern.
                if re.match(f"^{include:s}$", value)
            }
        )

        return terms_aggs

    def get_form_fields(self):
        """
        Indexables are filtered via:
        - a list of their Elasticsearch ids i.e strings,
        - a regex to eventually limit which terms are facetted.
        """
        return {
            self.name: (
                ArrayField(required=False, base_type=forms.CharField(max_length=50)),
                True,  # an ArrayField expects list values
            ),
            f"{self.name:s}_include": (
                forms.CharField(max_length=500, required=False),
                False,  # a CharField expects string values
            ),
        }

    def get_i18n_names(self, keys):
        """
        Helper method to get the corresponding internationalized human name for each key in
        a list of indexed objects' ids.
        This covers the base case for terms e.g. other models in their own ElasticSearch index
        like organizations or categories.
        """
        language = translation.get_language()
        indexer = getattr(ES_INDICES, self.term)

        # Get just the documents we need from ElasticSearch
        # pylint: disable=unexpected-keyword-arg
        search_query_response = ES_CLIENT.search(
            # We only need the titles to get the i18n names
            _source=["title"],
            index=indexer.index_name,
            body={"query": {"terms": {"_id": keys}}},
            size=len(keys),
        )

        # Extract the best available language here to avoid handling these kinds of
        # implementation details in the ViewSet
        return {
            doc["_id"]: get_best_field_language(doc["_source"]["title"], language)
            for doc in search_query_response["hits"]["hits"]
        }

    def get_static_definitions(self):
        """
        Build the static definition from the filter's base properties.
        """
        return {
            self.name: {
                "base_path": self.base_page.node.path if self.base_page else None,
                "human_name": self.human_name,
                "is_autocompletable": self.is_autocompletable,
                "is_drilldown": self.is_drilldown,
                "is_searchable": self.is_searchable,
                "name": self.name,
                "position": self.position,
            }
        }

    def get_faceted_definitions(self, facets, data, *args, **kwargs):
        """
        Build the filter definition's values from base definition and the faceted keys in the
        current language.
        Those provide us with the keys and counts that we just have to consume. They come from:
        - a bucket with the top facets in decreasing order of counts,
        - specific facets for values that were select in the querystring (we must force them
          because they may not be in the n top facet counts but we must make sure we keep it
          so that it remains available as an option so the user sees it and can unselect it)

        We resort to the `get_i18n_names` method to get the internationalized human names.
        """
        # Convert the keys & counts from ElasticSearch facets to a more readily consumable format
        #   {
        #       self.name: {
        #           self.name: {
        #               "buckets": [
        #                   {"key": "A", "count": x}
        #                   {"key": "B", "count": y}
        #                   {"key": "C", "count": z}
        #               ]
        #           }
        #       }
        #   }
        # 👆 becomes 👇
        #   [
        #       ("A", x),
        #       ("B", y),
        #       ("C", z),
        #   ]
        # sorted from highest facet count to lowest.
        key_count_pairs = sorted(
            [
                (key_count_dict["key"], key_count_dict["doc_count"])
                for key_count_dict in facets[self.name][self.name]["buckets"]
            ],
            key=itemgetter(1),
            reverse=True,
        )

        # We requested more facets than needed to be able to determine with certainty if there
        # are more available than what we're returning (see in `get_aggs_fragment`).
        base_facet_limit = applicable_facet_limit(data, self.name)
        # There are four possible outcomes:
        #   - We have as many results as we requested: there are more values (at least the `+1` we
        #     added to our request besides the base limit and the active values);
        #   - We have no value(s) besides the base limit: we can safely conclude there are no more
        #     values (eg. we did not get that requested `+1`)
        #   - We are somewhere between one more than the base limit and the requested number: this
        #     is ambiguous because it depends on the placement of our active values (which are
        #     always included no matter the circumstances).
        #       => If at least one of the values beyond base limit is not an active value, we know
        #          we're not going to show it. Therefore there are more values.
        #       => Otherwise, meaning if all the values beyond the base limit happen to be our
        #          active values, we are going to show all of them and therefore there is nothing
        #          more to request.
        has_more_values = any(
            filter(
                lambda c: c[0] not in data[self.name],
                key_count_pairs[base_facet_limit:],
            )
        )

        # Now that we've used our extra values to determine if there are more values to be fetched,
        # we can trim our counts down to the top N, based on the applicable limit.
        key_count_map = dict(key_count_pairs[:base_facet_limit])

        # Add the facets that were forced because their value was selected in querystring
        # Their result is of the form: {'organizations@P-0001': {'doc_count': 12}}
        key_count_map.update(
            {
                key.split("@")[1]: facet["doc_count"]
                for key, facet in facets.items()
                if f"{self.name:s}@" in key  # 6 times faster than startswith
            }
        )

        # Get internationalized names for all our keys
        key_i18n_name_map = self.get_i18n_names([*key_count_map])

        # Add human names to keys and counts before sorting as some of our sortings
        # use alphabetical ordering.
        values = [
            # Aggregate the information from right above to build the values
            {
                "count": count,
                "human_name": key_i18n_name_map[key],
                "key": key,
            }
            for key, count in key_count_map.items()
        ]

        # Sort facets as requested
        if self.sorting == self.SORTING_NAME:
            # Alphabetical ascending sorting
            values = sorted(
                values,
                key=lambda value: (value["human_name"], value["count"] * -1),
            )
        elif self.sorting == self.SORTING_COUNT:
            # Sorting by descending facet count
            values = sorted(
                values,
                key=lambda value: (value["count"] * -1, value["human_name"]),
            )
        else:
            # NB: self.SORTING_CONF is not appropriate for Indexables as they are not defined
            # in the static filter definition but generated from indices.
            raise ImproperlyConfigured(
                f'Facet sorting "{self.sorting}" is invalid for filter {self.name}.'
            )

        return {
            self.name: {
                # We always need to pass the base definition to the frontend
                **self.get_static_definitions()[self.name],
                # If values are removed due to `min_doc_count`, we are not returning them, and
                # therefore by definition our filter `has_more_values`.
                "has_more_values": has_more_values
                or any(value["count"] < self.min_doc_count for value in values),
                "values": list(
                    filter(
                        # We filter out values that do not meet `min_doc_count`manually instead of
                        # using ElasticSearch's builtin feature so it does not interfere with our
                        # `has_more_values` check.
                        lambda value: value["count"] >= self.min_doc_count
                        or value["key"] in data[self.name],
                        values,
                    )
                ),
            }
        }


class IndexableMPTTFilterDefinition(IndexableFilterDefinition):
    """
    Some of our filters are a special case of terms-based filter as they use MPTT paths
    as IDs and enable limiting faceting by path.
    """

    @property
    def aggs_include(self):
        """
        Return a regex that limits what facets are computed on the field.

        Returns:
        --------
            string: a regex depending on filters configuration in settings and pages in the CMS:
            - "": the empty string, if the `reverse_id` does not correspond to any published
                page which will not match any value and return an empty list of facets,
            - ".*-0001.{4}": if the `reverse_id` points to a published page with path "0001"
                this will match ids of the children of this page that will be of the form
                P-00010001 if they have children or L-00010001 if they are leafs,
            ' ".*": if no `reverse_id` is set (delegated to super) which will
                match all values.
        """
        if self.reverse_id:
            if self.base_page:
                node = self.base_page.node
                return f".*-{node.path:s}.{{{node.steplen:d}}}"
            return ""

        return super().aggs_include


class StaticChoicesFilterDefinition(
    ChoicesAggsMixin, ChoicesQueryMixin, BaseChoicesFilterDefinition
):
    """
    A filter definition for static choices ie that can be defined from the project settings.
    """

    def __init__(self, values, fragment_map, *args, **kwargs):
        """Record values and fragment map as attributes."""
        super().__init__(*args, **kwargs)
        self.values = values
        self.fragment_map = fragment_map

    def get_values(self):
        """Return the values recorded as attribute."""
        return self.values

    def get_fragment_map(self):
        """Return the fragment map recorded as attribute."""
        return self.fragment_map


class AvailabilityFilterDefinition(
    NestedChoicesAggsMixin, ChoicesQueryMixin, BaseChoicesFilterDefinition
):
    """
    Filter definition to allow filtering by availability. Hardcoded choices provided along with
    their Elasticsearch filtering fragment.
    """

    COMING_SOON, OPEN, ONGOING, ARCHIVED = "coming_soon", "open", "ongoing", "archived"

    def get_values(self):
        """Return the hardcoded values with internationalized human names."""
        return {
            self.OPEN: _("Open for enrollment"),
            self.COMING_SOON: _("Coming soon"),
            self.ONGOING: _("On-going"),
            self.ARCHIVED: _("Archived"),
        }

    def get_fragment_map(self):
        """Return the hardcoded query fragments updated to the current datetime."""
        return {
            self.OPEN: [
                {"range": {"course_runs.enrollment_start": {"lte": timezone.now()}}},
                {"range": {"course_runs.enrollment_end": {"gte": timezone.now()}}},
            ],
            self.COMING_SOON: [
                {"range": {"course_runs.start": {"gte": timezone.now()}}}
            ],
            self.ONGOING: [
                {"range": {"course_runs.start": {"lte": timezone.now()}}},
                {"range": {"course_runs.end": {"gte": timezone.now()}}},
            ],
            self.ARCHIVED: [{"range": {"course_runs.end": {"lte": timezone.now()}}}],
        }


class LanguagesFilterDefinition(
    NestedChoicesAggsMixin, TermsQueryMixin, BaseChoicesFilterDefinition
):
    """
    Languages need their own FilterDefinition subclass as there's a different way to get their
    human names from other terms-based filters and the filter is applied on the `course_runs`
    nested field.
    """

    def get_values(self):
        """Return the language values defined in the project's settings."""
        return ALL_LANGUAGES_DICT

    def get_fragment_map(self):
        """Compute query fragments for each language defined in the project's settings."""
        return {
            language: [{"term": {"course_runs.languages": language}}]
            for language in ALL_LANGUAGES_DICT
        }



# Generated by Django 3.1.4 on 2020-12-16 06:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("courses", "0023_auto_20201202_1146"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="blogpost",
            options={
                "ordering": ["-pk"],
                "verbose_name": "blog post",
                "verbose_name_plural": "blog posts",
            },
        ),
        migrations.AlterModelOptions(
            name="blogpostpluginmodel",
            options={
                "verbose_name": "blog post plugin",
                "verbose_name_plural": "blog post plugins",
            },
        ),
        migrations.AlterModelOptions(
            name="category",
            options={
                "ordering": ["-pk"],
                "verbose_name": "category",
                "verbose_name_plural": "categories",
            },
        ),
        migrations.AlterModelOptions(
            name="categorypluginmodel",
            options={
                "verbose_name": "category plugin",
                "verbose_name_plural": "category plugins",
            },
        ),
        migrations.AlterModelOptions(
            name="course",
            options={"verbose_name": "course", "verbose_name_plural": "courses"},
        ),
        migrations.AlterModelOptions(
            name="coursepluginmodel",
            options={
                "verbose_name": "course plugin",
                "verbose_name_plural": "course plugins",
            },
        ),
        migrations.AlterModelOptions(
            name="courserun",
            options={
                "verbose_name": "course run",
                "verbose_name_plural": "course runs",
            },
        ),
        migrations.AlterModelOptions(
            name="courseruntranslation",
            options={
                "verbose_name": "Course run translation",
                "verbose_name_plural": "Course run translations",
            },
        ),
        migrations.AlterModelOptions(
            name="licence",
            options={"verbose_name": "licence", "verbose_name_plural": "licences"},
        ),
        migrations.AlterModelOptions(
            name="licencepluginmodel",
            options={
                "verbose_name": "licence plugin",
                "verbose_name_plural": "licence plugins",
            },
        ),
        migrations.AlterModelOptions(
            name="licencetranslation",
            options={
                "verbose_name": "Licence translation",
                "verbose_name_plural": "licence translations",
            },
        ),
        migrations.AlterModelOptions(
            name="organization",
            options={
                "ordering": ["-pk"],
                "verbose_name": "organization",
                "verbose_name_plural": "organizations",
            },
        ),
        migrations.AlterModelOptions(
            name="organizationpluginmodel",
            options={
                "verbose_name": "organization plugin",
                "verbose_name_plural": "organization plugins",
            },
        ),
        migrations.AlterModelOptions(
            name="organizationsbycategorypluginmodel",
            options={
                "verbose_name": "organizations by category plugin",
                "verbose_name_plural": "organizations by category plugins",
            },
        ),
        migrations.AlterModelOptions(
            name="pagerole",
            options={"verbose_name": "page role", "verbose_name_plural": "page roles"},
        ),
        migrations.AlterModelOptions(
            name="person",
            options={"verbose_name": "person", "verbose_name_plural": "persons"},
        ),
        migrations.AlterModelOptions(
            name="personpluginmodel",
            options={
                "verbose_name": "person plugin",
                "verbose_name_plural": "person plugins",
            },
        ),
        migrations.AlterModelOptions(
            name="program",
            options={
                "ordering": ["-pk"],
                "verbose_name": "program",
                "verbose_name_plural": "programs",
            },
        ),
        migrations.AlterModelOptions(
            name="programpluginmodel",
            options={
                "verbose_name": "program plugin",
                "verbose_name_plural": "program plugins",
            },
        ),
        migrations.AddField(
            model_name="course",
            name="code",
            field=models.CharField(
                blank=True,
                help_text="Unique course code",
                max_length=100,
                null=True,
                verbose_name="code",
            ),
        ),
    ]



"""LMS handler to select and return the right LMS backend for each url."""
import re

from django.conf import settings
from django.utils.module_loading import import_string


class LMSHandler:
    """Class to handle LMS backends.

    Actions on a particular course are automatically routed to the LMS handling this course
    via the `COURSE_REGEX` configured for each LMS.
    """

    @staticmethod
    def select_lms(url):
        """
        Select and return the first LMS backend matching the url passed in argument.

        Default to None if no LMS is found to enable use-cases where we need to detect whether
        a course run has a matching LMS or not. Callers can determine if not finding an LMS
        backend is an exception or not.
        """
        if url is None:
            return None

        for lms_configuration in settings.RICHIE_LMS_BACKENDS:
            if re.match(lms_configuration.get("COURSE_REGEX", r".*"), url):
                return import_string(lms_configuration["BACKEND"])(lms_configuration)

        return None



"""Declare API endpoints for LTI Consumer Plugin"""
from django.core.cache import caches
from django.core.cache.backends.base import InvalidCacheBackendError
from django.http import JsonResponse
from django.shortcuts import get_object_or_404

from rest_framework import viewsets
from rest_framework.decorators import action

from . import models


class LTIConsumerViewsSet(viewsets.GenericViewSet):
    """Viewset for LTI Consumer Plugin"""

    @action(methods=["get"], detail=True, url_path="context")
    # pylint: disable=no-self-use,unused-argument,invalid-name
    def get_context(self, request, version=None, pk=None):
        """Process context data for the LTI Consumer Plugin.

        Parameters:
        - pk: the primary key of the LTI Consumer plugin to get context

        Returns:
        - response (JSON Object):
            - url: the LTI resource url
            - content_paramters: all generated parameters related to the lti provider
            - is_automatic_resizing: boolean to control automatic resizing

        """
        cache_key = f"lti_consumer_plugin__pk_{pk}"
        edit = request.toolbar and request.toolbar.edit_mode_active

        # Send response from cache only if edition is off
        if edit:
            cache = None
        else:
            try:
                cache = caches["memory_cache"]
            except InvalidCacheBackendError:
                cache = None
            else:
                response = cache.get(cache_key)
                if response is not None:
                    return JsonResponse(response)

        plugin = get_object_or_404(models.LTIConsumer, pk=pk)

        # If edition is on, check permissions to make sure it is also allowed
        # before granting the instructor role
        edit = edit and plugin.placeholder.has_change_plugin_permission(
            request.user, plugin
        )

        response = {
            "is_automatic_resizing": plugin.lti_provider.get(
                "is_automatic_resizing", True
            )
            if plugin.lti_provider_id
            else plugin.is_automatic_resizing,
            "content_parameters": plugin.get_content_parameters(edit=edit),
            "url": plugin.url,
        }

        if cache is not None:
            # Cache the response for 5 minutes,
            # lti oauth credentials are stale after this delay.
            cache.set(cache_key, response, 5 * 60)

        return JsonResponse(response)



# Generated by Django 3.1.4 on 2020-12-16 08:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("courses", "0024_auto_20201216_0740"),
    ]

    operations = [
        migrations.AlterField(
            model_name="courseruntranslation",
            name="title",
            field=models.CharField(
                blank=True, max_length=255, null=True, verbose_name="title"
            ),
        ),
    ]



"""Elasticsearch indices for Richie's search app."""
from django.conf import settings

from ..utils.indexers import IndicesList

ES_INDICES = IndicesList(
    **getattr(
        settings,
        "ES_INDICES",
        {
            "categories": "richie.apps.search.indexers.categories.CategoriesIndexer",
            "courses": "richie.apps.search.indexers.courses.CoursesIndexer",
            "organizations": "richie.apps.search.indexers.organizations.OrganizationsIndexer",
            "persons": "richie.apps.search.indexers.persons.PersonsIndexer",
        },
    )
)



"""
ElasticSearch course document management utilities
"""
from collections import defaultdict
from datetime import datetime
from functools import reduce

from django.conf import settings
from django.db.models import Prefetch
from django.utils import translation

from cms.models import Title
from djangocms_picture.models import Picture

from richie.plugins.plain_text.models import PlainText
from richie.plugins.simple_picture.helpers import get_picture_info
from richie.plugins.simple_text_ckeditor.models import SimpleText

from ...courses.models import MAX_DATE, CategoryPluginModel, Course, CourseState
from ..defaults import ES_INDICES_PREFIX, ES_STATE_WEIGHTS
from ..forms import CourseSearchForm
from ..text_indexing import MULTILINGUAL_TEXT
from ..utils.i18n import get_best_field_language
from ..utils.indexers import slice_string_for_completion

BEST_STATE_SCRIPT = """
    DateTimeFormatter formatter = DateTimeFormatter.ofPattern(
        "yyyy-MM-dd'T'HH:mm:ss[.SSSSSS]XXXXX"
    );
    int best_state = 7;
    int best_index = 0;
    long start, end, enrollment_start, enrollment_end;
    Set intersection;

    // Go through the sorted course runs nested under this course to look for the
    // best course run (open for enrollment > future > on-going > archived)
    for (int i = 0; i < params._source.course_runs.length; ++i) {
        start = ZonedDateTime.parse(
            params._source.course_runs[i]['start'], formatter
        ).toInstant().toEpochMilli();
        end = ZonedDateTime.parse(
            params._source.course_runs[i]['end'], formatter
        ).toInstant().toEpochMilli();
        enrollment_start = ZonedDateTime.parse(
            params._source.course_runs[i]['enrollment_start'], formatter
        ).toInstant().toEpochMilli();
        enrollment_end = ZonedDateTime.parse(
            params._source.course_runs[i]['enrollment_end'], formatter
        ).toInstant().toEpochMilli();

        // Use language sets to check their intersection
        intersection = new HashSet(params._source.course_runs[i]['languages']);
        if (params.languages != null) {
            intersection.retainAll(new HashSet(params.languages))
        }

        // Only consider this course run if it is in a desired language
        if (params.languages == null || !intersection.isEmpty()) {
            if (start < params.ms_since_epoch) {
                if (end > params.ms_since_epoch) {
                    if (enrollment_end > params.ms_since_epoch) {
                        // on-going open
                        if (params.states == null || params.states.contains(0)) {
                            best_state = 0;
                            best_index = i;
                            // We already found the best... let's save iterations
                            break;
                        }
                    }
                    else {
                        // on-going closed
                        if (
                            best_state > 5 && (
                                params.states == null ||
                                params.states.contains(5)
                            )
                        ) {
                            best_state = 5;
                            best_index = i;
                        }
                    }
                }
                else {
                    if (enrollment_end > params.ms_since_epoch) {
                        // archived open
                        if (
                            best_state > 2 && (
                                params.states == null ||
                                params.states.contains(2)
                            )
                        ) {
                            best_state = 2;
                            best_index = i;
                            // Course runs are ordered by `end` date so we know all
                            // remaining course runs will not be better than
                            // archived open.
                            // Let's break to save iterations:
                            break;
                        }
                    }
                    else {
                        // archived closed
                        if (
                            best_state > 6 && (
                                params.states == null ||
                                params.states.contains(6)
                            )
                        ) {
                            best_state = 6;
                            best_index = i;
                        }
                    }
                }
            }
            else if (enrollment_start > params.ms_since_epoch) {
                // future not yet open
                if (
                    best_state > 3 && (
                        params.states == null ||
                        params.states.contains(3)
                    )
                ) {
                    best_state = 3;
                    best_index = i;
                }
            }
            else if (enrollment_end > params.ms_since_epoch) {
                // future open
                if (
                    best_state > 1 && (
                        params.states == null ||
                        params.states.contains(1)
                    )
                ) {
                    best_state = 1;
                    best_index = i;
                }
            }
            else {
                // future already closed
                if (
                    best_state > 4 && (
                        params.states == null ||
                        params.states.contains(4)
                    )
                ) {
                    best_state = 4;
                    best_index = i;
                }
            }
        }
    }

"""


class CoursesIndexer:
    """
    Makes available the parameters the indexer requires as well as functions to shape
    objects getting into and out of ElasticSearch
    """

    index_name = f"{ES_INDICES_PREFIX}_courses"
    mapping = {
        "dynamic_templates": MULTILINGUAL_TEXT,
        "properties": {
            "course_runs": {
                "type": "nested",
                "properties": {
                    "id": {"type": "keyword"},
                    "start": {"type": "date"},
                    "end": {"type": "date"},
                    "enrollment_start": {"type": "date"},
                    "enrollment_end": {"type": "date"},
                    "languages": {"type": "keyword"},
                },
            },
            # Keywords
            "categories": {"type": "keyword"},
            "organizations": {"type": "keyword"},
            "persons": {"type": "keyword"},
            # Searchable
            # description, title, category names & organization names are handled
            # by `MULTILINGUAL_TEXT`
            "code": {"type": "text", "analyzer": "code_trigram"},
            **{
                f"complete.{lang:s}": {
                    "type": "completion",
                    "analyzer": "simple_diacritics_insensitive",
                }
                for lang, _ in settings.LANGUAGES
            },
            # `persons_names` cannot be handled by `MULTILINGUAL_TEXT` because language
            # analyzers fail on people's names.
            **{
                "persons_names.{lang}": {"type": "text", "analyzer": "simple"}
                for lang, _ in settings.LANGUAGES
            },
            "is_new": {"type": "boolean"},
            "is_listed": {"type": "boolean"},
            # Not searchable
            "absolute_url": {"type": "object", "enabled": False},
            "cover_image": {"type": "object", "enabled": False},
            "icon": {"type": "object", "enabled": False},
        },
    }
    display_fields = [
        "absolute_url",
        "categories",
        "code",
        "cover_image",
        "duration",
        "effort",
        "icon",
        "introduction",
        "organization_highlighted",
        "organizations",
        "title",
    ]
    form = CourseSearchForm

    scripts = {
        # The ordering process first splits the courses into 7 buckets (Bucket 0 to Bucket 6),
        # based on their state with further ordering inside each one of those buckets.
        #
        # The priority of each bucket is determined by a weight coming from the settings:
        #     > RICHIE_ES_STATE_WEIGHTS = [80, 40, 20, 10, 6, 5, 1]
        #
        # The priority of each course within a bucket is determined by a delta that is based on
        # one of the course's datetimes (start, end or end_of_enrollment date depending on its
        # state). This delta is added for dates later than the current datetime (in order to rank
        # them by ascending order), and substracted for dates earlier than the current time (in
        # order to rank them by descending order). Said differently, courses are ranked within a
        # bucket by the descending proximity of their signifiant datetime with the current
        # datetime.
        #
        # The boost factor for each course is given by the formula
        # (We add 1 to _score so that it is always > 0 and never brings boost factors to 0):
        #
        #     > boost_factor = (_score + 1) * (weight * now -/+ significant_datetime)
        #
        # Here's a schematic representation that shows the ordering factor for each
        # course (to be used in descending order later on) :
        #
        #                   TOP OF THE LIST
        #                   ———————————————
        # --------------------- weight_0 ---------------------
        #     Bucket 0: Courses that are on-going and open
        #      sorted by ascending end of enrollment date
        #
        # --------------------- weight_1 ---------------------
        #      Bucket 1: Courses that are future and open
        #            sorted by ascending start date
        #
        # --------------------- weight_2 ---------------------
        #      Bucket 2: Courses that are archived but open
        #       sorted by ascending end of enrollment date
        #
        # --------------------- weight_3 ---------------------
        #  Bucket 3: Courses that are future but not yet open
        #            sorted by ascending start date
        #
        # --------------------- weight_4 ---------------------
        # Bucket 4: Courses that are future but already closed
        #            sorted by ascending start date
        #
        # --------------------- weight_5 ---------------------
        #     Bucket 5: Courses that are on-going but closed
        #           sorted by ascending end date
        #
        # --------------------- weight_6 ---------------------
        #          Bucket 6: Courses that are archived
        #            sorted by descending end date
        #
        # ------------------ Weight set to 0 -----------------
        #      Bucket 7: Courses that have no course runs
        #                No specific ordering
        #                   ———————————————
        #                   END OF THE LIST
        #
        "score": {
            "script": {
                "lang": "painless",
                "source": BEST_STATE_SCRIPT
                + """
                if (best_state == 0) {{
                    // The course is on-going and open
                    // Ordered by ascending end of enrollment datetime. The next course to
                    // end enrollment is displayed first.
                    return (_score + 1) * (
                        {weight_0:d} * params.ms_since_epoch - ZonedDateTime.parse(
                            params._source.course_runs[best_index]['enrollment_end'], formatter
                        ).toInstant().toEpochMilli()
                    );
                }}
                else if (best_state == 1) {{
                    // The course is future and open
                    // Ordered by starting datetime. The next course to start is displayed
                    // first.
                    return (_score + 1) * (
                        {weight_1:d} * params.ms_since_epoch - ZonedDateTime.parse(
                            params._source.course_runs[best_index]['start'], formatter
                        ).toInstant().toEpochMilli()
                    );
                }}
                else if (best_state == 2) {{
                    // The course is archived but open for enrollment
                    // Ordered by ascending end of enrollment datetime. The next course to
                    // end enrollment is displayed first.
                    return (_score + 1) * (
                        {weight_2:d} * params.ms_since_epoch - ZonedDateTime.parse(
                            params._source.course_runs[best_index]['enrollment_end'], formatter
                        ).toInstant().toEpochMilli()
                    );
                }}
                else if (best_state == 3) {{
                    // The course is future but not yet open for enrollment
                    // Ordered by starting datetime. The next course to start is displayed
                    // first.
                    return (_score + 1) * (
                        {weight_3:d} * params.ms_since_epoch - ZonedDateTime.parse(
                            params._source.course_runs[best_index]['start'], formatter
                        ).toInstant().toEpochMilli()
                    );
                }}
                else if (best_state == 4) {{
                    // The course is future but already closed for enrollment
                    // Ordered by end datetime. The next course to end is displayed first
                    return (_score + 1) * (
                        {weight_4:d} * params.ms_since_epoch - ZonedDateTime.parse(
                            params._source.course_runs[best_index]['start'], formatter
                        ).toInstant().toEpochMilli()
                    );
                }}
                else if (best_state == 5) {{
                    // The course is on-going and closed for enrollment
                    // Ordered by end datetime. The next course to end is displayed first
                    return (_score + 1) * (
                        {weight_5:d} * params.ms_since_epoch - ZonedDateTime.parse(
                            params._source.course_runs[best_index]['end'], formatter
                        ).toInstant().toEpochMilli()
                    );
                }}
                else if (best_state == 6) {{
                    // The course is archived and closed for enrollment
                    // Ordered by end datetime. The next course to start is displayed
                    // first.
                    return (_score + 1) * (
                        {weight_6:d} * params.ms_since_epoch + ZonedDateTime.parse(
                            params._source.course_runs[best_index]['end'], formatter
                        ).toInstant().toEpochMilli()
                    );
                }}
                // The course has no course runs
                return 0;
                """.format(  # pylint: disable=consider-using-f-string
                    **{
                        f"weight_{i:d}": weight
                        for i, weight in enumerate(ES_STATE_WEIGHTS)
                    }
                ),
            }
        },
        # Compute a course's state based on the current state of each of its course runs.
        "state_field": {
            "script": {
                "lang": "painless",
                "source": BEST_STATE_SCRIPT
                + """
                if (best_state == 0) {
                    // The course is on-going and open
                    // Return the date of end of enrollment
                    return [
                        'priority': 0,
                        'date_time': params._source.course_runs[best_index][
                            'enrollment_end']
                    ];
                }
                else if (best_state == 1) {
                    // The course is future and open
                    // Return the start date
                    return [
                        'priority': 1,
                        'date_time': params._source.course_runs[best_index]['start']
                    ];
                }
                else if (best_state == 2) {
                    // The course is archived but open
                    // Return the date of end of enrollment
                    return [
                        'priority': 2,
                        'date_time': params._source.course_runs[best_index][
                            'enrollment_end']
                    ];
                }
                else if (best_state == 3) {
                    // The course is future but not yet open for enrollment
                    // Return the start date
                    return [
                        'priority': 3,
                        'date_time': params._source.course_runs[best_index]['start']
                    ];
                }
                else if (best_state == 4) {
                    // The course is future but already closed for enrollment
                    return ['priority': 4];
                }
                else if (best_state == 5) {
                    // The course is on-going and closed for enrollment
                    return ['priority': 5];
                }
                else if (best_state == 6) {
                    // The course is archived
                    return ['priority': 6];
                }
                else {
                    // The course has no course runs
                    return ['priority': 7];
                }
                """,
            }
        },
    }

    # pylint: disable=too-many-locals
    @classmethod
    def get_es_document_for_course(cls, course, index=None, action="index"):
        """
        Build an Elasticsearch document from the course instance.
        """
        index = index or cls.index_name

        # Prepare published titles
        titles = {
            t.language: t.title
            for t in Title.objects.filter(page=course.extended_object, published=True)
        }

        # Prepare cover images
        cover_images = {}
        for cover in Picture.objects.filter(
            cmsplugin_ptr__placeholder__page=course.extended_object,
            cmsplugin_ptr__placeholder__slot="course_cover",
        ):
            language = cover.cmsplugin_ptr.language
            with translation.override(language):
                picture_info = get_picture_info(cover, "cover")
                if picture_info:
                    cover_images[language] = picture_info

        # Prepare the related category icon
        icon_images = {}
        for plugin_model in CategoryPluginModel.objects.filter(
            cmsplugin_ptr__placeholder__page=course.extended_object_id,
            cmsplugin_ptr__placeholder__slot="course_icons",
            cmsplugin_ptr__position=0,
        ):
            language = plugin_model.language
            for icon in Picture.objects.filter(
                cmsplugin_ptr__language=language,
                cmsplugin_ptr__placeholder__page=plugin_model.page_id,
                cmsplugin_ptr__placeholder__slot="icon",
                cmsplugin_ptr__position=0,
            ):
                with translation.override(language):
                    picture_info = get_picture_info(icon, "icon") or {}
                    icon_images[language] = {
                        **picture_info,
                        "color": plugin_model.page.category.color,
                        "title": plugin_model.page.get_title(),
                    }

        # Prepare description texts
        descriptions = defaultdict(list)
        for simple_text in SimpleText.objects.filter(
            cmsplugin_ptr__placeholder__page=course.extended_object,
            cmsplugin_ptr__placeholder__slot="course_description",
        ):
            descriptions[simple_text.cmsplugin_ptr.language].append(simple_text.body)

        # Prepare introduction texts
        introductions = defaultdict(list)
        for plain_text in PlainText.objects.filter(
            cmsplugin_ptr__placeholder__page=course.extended_object,
            cmsplugin_ptr__placeholder__slot="course_introduction",
        ):
            introductions[plain_text.cmsplugin_ptr.language].append(plain_text.body)

        # Prepare localized duration texts
        duration = {}
        for language, _ in settings.LANGUAGES:
            with translation.override(language):
                duration[language] = course.get_duration_display()

        # Prepare localized effort texts
        effort = {}
        for language, _ in settings.LANGUAGES:
            with translation.override(language):
                effort[language] = course.get_effort_display()

        # Prepare categories, making sure we get title information for categories
        # in the same query
        category_pages = (
            course.get_root_to_leaf_public_category_pages()
            .select_related("node")
            .prefetch_related(
                Prefetch(
                    "title_set",
                    to_attr="published_titles",
                    queryset=Title.objects.filter(published=True),
                )
            )
            .only("node", "pk")
            .distinct()
        )

        # Prepare organizations, making sure we get title information for organizations
        # in the same query
        organizations = (
            course.get_organizations()
            .prefetch_related(
                Prefetch(
                    "extended_object__title_set",
                    to_attr="published_titles",
                    queryset=Title.objects.filter(published=True),
                )
            )
            .only("extended_object__node")
            .distinct()
        )
        organization_main = course.get_main_organization()
        organization_highlighted = (
            organizations.get(id=organization_main.id) if organization_main else None
        )

        # Prepare persons, making sure we get title information for persons
        # in the same query
        persons = (
            course.get_persons()
            .prefetch_related(
                Prefetch(
                    "extended_object__title_set",
                    to_attr="published_titles",
                    queryset=Title.objects.filter(published=True),
                )
            )
            .distinct()
        )

        # Prepare course runs
        # Ordering them by their `end` date is important to optimize sorting and other
        # computations that require looping on the course runs
        # Course runs with no start date or no start of enrollment date are ignored as
        # they are still to be scheduled.
        course_runs = [
            {
                "start": cr["start"],
                "end": cr["end"] or MAX_DATE,
                "enrollment_start": cr["enrollment_start"],
                "enrollment_end": cr["enrollment_end"] or cr["end"] or MAX_DATE,
                "languages": cr["languages"],
            }
            for cr in course.course_runs.filter(
                start__isnull=False, enrollment_start__isnull=False
            )
            .order_by("-end")
            .values("start", "end", "enrollment_start", "enrollment_end", "languages")
        ]

        return {
            "_id": str(course.extended_object_id),
            "_index": index,
            "_op_type": action,
            "absolute_url": {
                lang: course.extended_object.get_absolute_url(lang)
                for lang, _ in settings.LANGUAGES
            },
            "categories": [page.category.get_es_id() for page in category_pages],
            # Index the names of categories to surface them in full text searches
            "categories_names": reduce(
                lambda acc, title: {
                    **acc,
                    title.language: acc[title.language] + [title.title]
                    if acc.get(title.language)
                    else [title.title],
                },
                [title for page in category_pages for title in page.published_titles],
                {},
            ),
            "code": course.code,
            "complete": {
                language: slice_string_for_completion(title)
                for language, title in titles.items()
            }
            if course.is_listed
            else None,
            "course_runs": course_runs,
            "cover_image": cover_images,
            "description": {
                language: " ".join(st) for language, st in descriptions.items()
            },
            "duration": duration,
            "effort": effort,
            "icon": icon_images,
            "introduction": {
                language: " ".join(st) for language, st in introductions.items()
            },
            "is_new": len(course_runs) == 1,
            # If titles is an empty dict, it means the course is not published in any language:
            "is_listed": bool(course.is_listed and titles),
            # Pick the highlighted organization from the organizations QuerySet to benefit from
            # the prefetch of related title sets
            "organization_highlighted": {
                title.language: title.title
                for title in organization_highlighted.extended_object.published_titles
            }
            if organization_highlighted
            else None,
            "organizations": [
                organization.get_es_id() for organization in organizations
            ],
            # Index the names of organizations to surface them in full text searches
            "organizations_names": reduce(
                lambda acc, title: {
                    **acc,
                    title.language: acc[title.language] + [title.title]
                    if acc.get(title.language)
                    else [title.title],
                },
                [
                    title
                    for organization in organizations
                    for title in organization.extended_object.published_titles
                ],
                {},
            ),
            "persons": [
                str(person.public_extension.extended_object_id) for person in persons
            ],
            "persons_names": reduce(
                lambda acc, title: {
                    **acc,
                    title.language: acc[title.language] + [title.title]
                    if acc.get(title.language)
                    else [title.title],
                },
                [
                    title
                    for person in persons
                    for title in person.extended_object.published_titles
                ],
                {},
            ),
            "title": titles,
        }

    @classmethod
    def get_es_documents(cls, index=None, action="index"):
        """
        Loop on all the courses in database and format them for the ElasticSearch index
        """
        index = index or cls.index_name

        for course in (
            Course.objects.filter(
                extended_object__publisher_is_draft=False,  # index the public object
                extended_object__title_set__published=True,  # only index published courses
                extended_object__node__parent__cms_pages__course__isnull=True,  # exclude snapshots
            )
            .distinct()
            .iterator()
        ):
            yield cls.get_es_document_for_course(course, index=index, action=action)

    @staticmethod
    def format_es_object_for_api(es_course, language=None):
        """
        Format a course stored in ES into a consistent and easy-to-consume record for
        API consumers
        """
        language = language or translation.get_language()
        source = es_course["_source"]

        # Prepare the state
        state = es_course["fields"]["state"][0]
        try:
            state["date_time"] = datetime.fromisoformat(state["date_time"])
        except KeyError:
            state["date_time"] = None

        return {
            **{
                field: get_best_field_language(source[field], language)
                for field in [
                    "absolute_url",
                    "cover_image",
                    "duration",
                    "effort",
                    "icon",
                    "introduction",
                    "title",
                ]
            },
            "id": es_course["_id"],
            "categories": source["categories"],
            "code": source["code"],
            "organization_highlighted": get_best_field_language(
                source["organization_highlighted"], language
            )
            if source.get("organization_highlighted", None)
            else None,
            "organizations": source["organizations"],
            "state": CourseState(**state),
        }

    @staticmethod
    def format_es_document_for_autocomplete(es_document, language=None):
        """
        Format a document stored in ES into an easy-to-consume record for autocomplete consumers.
        This method differs from the regular one as objects retrieved from query VS complete
        queries can be formatted differently; and consumers of autocomplete do not need
        full objects.
        """
        return {
            "absolute_url": get_best_field_language(
                es_document["_source"]["absolute_url"], language
            ),
            "id": es_document["_id"],
            "kind": "courses",
            "title": get_best_field_language(es_document["_source"]["title"], language),
        }
